import os
import shutil
import platform
import pickle
import json
import datetime
from multiprocessing import Pool

from tqdm import tqdm

try:
    from mpi4py import MPI
except:
    pass
import jpype

from .create_qseis2025 import (
    create_dir_qseis2025,
    create_inp_qseis2025,
    call_qseis2025,
    convert_pd2bin_qseis2025,
)
from .pytaup import create_tpts_table
from .utils import group, convert_earth_model_nd2nd_without_Q


def _call_qseis2025_star(args):
    return call_qseis2025(*args)


def pre_process_qseis2025(
        processes_num,
        path_green,
        path_bin,
        event_depth_list,
        receiver_depth_list,
        dist_range,
        delta_dist,
        N_each_group,
        time_window,
        sampling_interval,
        output_observables,
        slowness_int_algorithm=0,
        eps_estimate_wavenumber=1e-6,
        source_radius_ratio=0.05,
        slowness_window=None,
        time_reduction_velo=0,
        wavenumber_sampling_rate=12,
        anti_alias=0.01,
        free_surface=True,
        wavelet_duration=0,
        wavelet_type=1,
        flat_earth_transform=True,
        path_nd=None,
        earth_model_layer_num=None,
        check_finished_tpts_table=False,
):
    print("Preprocessing")
    os.makedirs(path_green, exist_ok=True)
    if platform.system() == "Windows":
        path_bin_call = os.path.join(path_green, "qseis2025.exe")
    else:
        path_bin_call = os.path.join(path_green, "qseis2025.bin")
    shutil.copy(path_bin, path_bin_call)

    N_dist, N_dist_group = None, None
    for event_depth in event_depth_list:
        for receiver_depth in receiver_depth_list:
            # print("creating green func dir, info, inp for event_depth=%.2f receiver_depth=%.2f" %
            #       (event_depth, receiver_depth))
            N_dist, N_dist_group = create_dir_qseis2025(
                path_green=path_green,
                event_depth=event_depth,
                receiver_depth=receiver_depth,
                dist_range=dist_range,
                delta_dist=delta_dist,
                N_each_group=N_each_group,
            )
            create_inp_qseis2025(
                path_green=path_green,
                event_depth=event_depth,
                receiver_depth=receiver_depth,
                dist_range=dist_range,
                delta_dist=delta_dist,
                N_dist=N_dist,
                N_dist_group=N_dist_group,
                N_each_group=N_each_group,
                time_window=time_window,
                sampling_interval=sampling_interval,
                output_observables=output_observables,
                slowness_int_algorithm=slowness_int_algorithm,
                eps_estimate_wavenumber=eps_estimate_wavenumber,
                source_radius_ratio=source_radius_ratio,
                slowness_window=slowness_window,
                time_reduction_velo=time_reduction_velo,
                wavenumber_sampling_rate=wavenumber_sampling_rate,
                anti_alias=anti_alias,
                free_surface=free_surface,
                wavelet_duration=wavelet_duration,
                wavelet_type=wavelet_type,
                flat_earth_transform=flat_earth_transform,
                path_nd=path_nd,
                earth_model_layer_num=earth_model_layer_num,
            )

    path_nd_without_Q = os.path.join(path_green, "noQ.nd")
    convert_earth_model_nd2nd_without_Q(path_nd, path_nd_without_Q)

    # creating tp and ts tables
    for event_depth in tqdm(event_depth_list, desc="Creating travel time tables"):
        for receiver_depth in receiver_depth_list:
            create_tpts_table(
                path_green,
                event_depth,
                receiver_depth,
                dist_range,
                delta_dist,
                path_nd_without_Q,
                check_finished_tpts_table,
            )
    if jpype.isJVMStarted():
        jpype.shutdownJVM()

    green_info = {
        "processes_num": processes_num,
        "event_depth_list": event_depth_list,
        "receiver_depth_list": receiver_depth_list,
        "dist_range": dist_range,
        "delta_dist": delta_dist,
        "N_dist": N_dist,
        "N_dist_group": N_dist_group,
        "N_each_group": N_each_group,
        "time_window": time_window,
        "sampling_interval": sampling_interval,
        "sampling_num": round(time_window / sampling_interval + 1),
        "slowness_int_algorithm": slowness_int_algorithm,
        "slowness_window": slowness_window,
        "time_reduction_velo": time_reduction_velo,
        "wavenumber_sampling_rate": wavenumber_sampling_rate,
        "anti_alias": anti_alias,
        "free_surface": free_surface,
        "wavelet_duration": wavelet_duration,
        "wavelet_type": wavelet_type,
        "flat_earth_transform": flat_earth_transform,
        "path_nd": path_nd,
        "path_nd_without_Q": path_nd_without_Q,
        "earth_model_layer_num": earth_model_layer_num,
    }
    json_str = json.dumps(green_info, indent=4, ensure_ascii=False)
    with open(
            os.path.join(path_green, "green_lib_info.json"), "w", encoding="utf-8"
    ) as file:
        file.write(json_str)

    inp_list = []
    for event_dep in event_depth_list:
        for receiver_dep in receiver_depth_list:
            for nn in range(N_dist_group):
                inp_list.append([event_dep, receiver_dep, nn])
    #inp_list_sorted = sorted(inp_list, key=lambda x: abs(x[0] - x[1]))
    group_list = group(inp_list, processes_num)
    with open(os.path.join(path_green, "group_list.pkl"), "wb") as fw:
        pickle.dump(group_list, fw)  # type: ignore
    return group_list


def create_grnlib_qseis2025_sequential(
        path_green, check_finished=False, convert_pd2bin=True, remove_pd=True
):
    with open(os.path.join(path_green, "group_list.pkl"), "rb") as fr:
        group_list = pickle.load(fr)
    for item in tqdm(group_list, desc="Computing dynamic stress"):
        for i in range(len(item)):
            # print("computing " + str(item[i]))
            item[i] = item[i] + [path_green, check_finished]
            call_qseis2025(*item[i])
    if convert_pd2bin:
        convert_pd2bin_qseis2025_all(path_green, remove_pd)


def create_grnlib_qseis2025_parallel(
        path_green, check_finished=False, convert_pd2bin=True, remove_pd=True
):
    with open(os.path.join(path_green, "group_list.pkl"), "rb") as fr:
        group_list = pickle.load(fr)
    tasks = []
    for grp in group_list:
        for item in grp:
            tasks.append(tuple(item + [path_green, check_finished]))

    processes = None
    try:
        with open(os.path.join(path_green, "green_lib_info.json"), "r") as fr:
            processes = json.load(fr).get("processes_num", None)
    except Exception:
        pass

    with Pool(processes=processes) as pool:
        for _ in tqdm(
                pool.imap_unordered(_call_qseis2025_star, tasks, chunksize=1),
                total=len(tasks),
                desc="Compute QSEIS2025 Green\'s library in the transformed domain.",
        ):
            pass

    if convert_pd2bin:
        convert_pd2bin_qseis2025_all(path_green, remove_pd)


def convert_pd2bin_qseis2025_all(path_green, remove=False):
    print("converting ascii files to bytes files")
    with open(os.path.join(path_green, "green_lib_info.json"), "r") as fr:
        green_info = json.load(fr)
    event_depth_list = green_info["event_depth_list"]
    receiver_depth_list = green_info["receiver_depth_list"]
    for event_dep in event_depth_list:
        for receiver_dep in receiver_depth_list:
            sub_dir = str(
                os.path.join(path_green, "%.2f" % event_dep, "%.2f" % receiver_dep)
            )
            sub_sub_dirs = os.listdir(sub_dir)
            for sub_sub_dir in sub_sub_dirs:
                if "_table.bin" not in sub_sub_dir:
                    convert_pd2bin_qseis2025(
                        os.path.join(sub_dir, sub_sub_dir), remove=remove
                    )


def create_grnlib_qseis2025_parallel_multi_nodes(
        path_green, check_finished=False, convert_pd2bin=True, remove_pd=True
):
    s = datetime.datetime.now()
    with open(os.path.join(path_green, "group_list.pkl"), "rb") as fr:
        group_list = pickle.load(fr)
    comm = MPI.COMM_WORLD
    processes_num = comm.Get_size()
    if processes_num != len(group_list[0]):
        raise ValueError(
            "processes_num is %d, item num in group is %d. \n"
            "Pleasse check the process num!" % (processes_num, len(group_list[0]))
        )
    rank = comm.Get_rank()
    ind_group = rank // processes_num
    ind_para = rank - ind_group * processes_num
    print(rank, ind_group, ind_para)
    call_qseis2025(
        event_depth=group_list[ind_group][ind_para][0],
        receiver_depth=group_list[ind_group][ind_para][1],
        n_group=group_list[ind_group][ind_para][2],
        path_green=path_green,
        check_finished=check_finished,
    )
    if convert_pd2bin:
        convert_pd2bin_qseis2025_all(path_green, remove_pd)
    e = datetime.datetime.now()
    print("run time:" + str(e - s))


if __name__ == "__main__":
    pass



import sys

from pygrnwang.focal_mechanism import plane2mt
from pygrnwang.geo import cal_first_p_s
from pygrnwang.read_qssp2020 import seek_qssp2020

from scipy.io import savemat
from scipy import signal
import numpy as np

if __name__ == "__main__":
    import time

    s = time.time()
    path_green = "/e/grns_test/test_qssp"
    event_depth_km = 10
    receiver_depth_km = 1
    az_deg = 60
    dist_km = 100
    focal_mechanism = [30, 40, 50]
    srate = 2
    before_p = None
    pad_zeros = True
    shift = False
    rotate = True
    only_seismograms = False
    output_type = "stress"
    model_name = "/e/grns_test/test_qssp/noq.nd"
    for i in range(1):
        seismograms, tpts_table, first_p, first_s, grn_dep, grn_receiver, green_dist = (
            seek_qssp2020(
                path_green,
                event_depth_km,
                receiver_depth_km,
                az_deg,
                dist_km,
                focal_mechanism,
                srate,
                before_p,
                pad_zeros,
                shift,
                rotate,
                only_seismograms,
                output_type,
                model_name,
            )
        )
    e = time.time()
    print("run time:", e - s, "s")
    #
    # for i in range(seismograms.shape[0]):
    #     w = signal.firwin(11, 0.1, fs=srate / 2)
    #     seismograms[i] = signal.convolve(seismograms[i], w, "same")
    #     # w = signal.windows.hann(srate)
    #     # seismograms[i] = signal.fftconvolve(seismograms[i],w, 'same')
    mu = 31126160000
    area = 1  # 1e6
    disp_source = 1
    m0 = mu * area * disp_source
    print(m0)
    print(seismograms.shape)
    print(seismograms[1, round(0.4 * dist_km * srate + 1)] * m0)
    print(tpts_table, first_p, first_s, grn_dep, grn_receiver, green_dist)

    first_p, _ = cal_first_p_s(
        event_depth_km, dist_km, receiver_depth_km,  "/e/grns_test/test_qssp/noq.nd"
    )

    import matplotlib.pyplot as plt

    t = np.linspace(0, len(seismograms[0]) / srate, len(seismograms[0]), endpoint=False)
    fig, axs = plt.subplots(6, 1)
    for ind in range(6):
        axs[ind].plot(t, seismograms[ind])
    axs[-1].set_xlabel("Time (s)")
    axs[-1].legend(loc="upper right")
    plt.show()


import os.path
import sys

import numpy as np
import pandas as pd

from scipy.signal import firwin
from pygrnwang.signal_process import filter_butter
from pygrnwang.read_qseis2025 import *
from pygrnwang.geo import d2km
from pygrnwang.read_by_qssp import read_by_qssp
from pygrnwang.read_qssp2020 import seek_qssp2020
from pygrnwang.read_edcmp import *
from pygrnwang.create_edcmp_bulk import *
from pygrnwang.signal_process import taper
from pygrnwang.pytaup import cal_first_p
from pygrnwang.focal_mechanism import *
import matplotlib
matplotlib.use('tkagg')
import matplotlib.pyplot as plt

if __name__ == "__main__":
    cut_length = 128
    put_pr = np.load('/home/zjc/Desktop/put_pr.npy')
    #put_pr_m = np.load('/home/zjc/Desktop/put_pr_m.npy')
    put_pr_diff = np.load('/home/zjc/Desktop/put_pr_diff.npy')

    plt.figure()
    plt.plot(put_pr[:cut_length], label='qseis_stress')
    #plt.plot(put_pr_m[:cut_length], label='qseis_stress_dj')
    plt.plot(put_pr_diff[:cut_length], label='qseis_stress_diff')
    plt.legend()
    plt.show()

    pur_pr = np.load('/home/zjc/Desktop/pur_pr.npy')
    #pur_pr_m = np.load('/home/zjc/Desktop/pur_pr_m.npy')
    pur_pr_diff = np.load('/home/zjc/Desktop/pur_pr_diff.npy')

    plt.figure()
    plt.plot(pur_pr[:cut_length], label='qseis_stress')
    #plt.plot(pur_pr_m[:cut_length], label='qseis_stress_dj')
    plt.plot(pur_pr_diff[:cut_length], label='qseis_stress_diff')
    plt.legend()
    plt.show()



import os
import hashlib
import subprocess
import platform
import json

import numpy as np
import pandas as pd

from .qssp2020inp import s as str_inp
from .signal_process import resample
from .utils import convert_earth_model_nd2inp
from .focal_mechanism import check_convert_fm, convert_mt_axis


def hash_read_pars(
    event_lat,
    event_lon,
    event_depth,
    receiver_lat,
    receiver_lon,
    receiver_depth,
):
    data = (
        str(float(event_lat)).encode()
        + str(float(event_lon)).encode()
        + str(float(event_depth)).encode()
        + str(float(receiver_lat)).encode()
        + str(float(receiver_lon)).encode()
        + str(float(receiver_depth)).encode()
    )
    return hashlib.md5(data).hexdigest()


def create_inp_qssp2020_read(
    read_name,
    path_green,
    event_lat,
    event_lon,
    event_depth,
    receiver_lat,
    receiver_lon,
    receiver_depth,
    focal_mechanism,
    green_info,
):
    spec_time_window = green_info["spec_time_window"]
    sampling_interval = green_info["sampling_interval"]
    max_frequency = green_info["max_frequency"]
    max_slowness = green_info["max_slowness"]
    anti_alias = green_info["anti_alias"]
    turning_point_filter = green_info["turning_point_filter"]
    turning_point_d1 = green_info["turning_point_d1"]
    turning_point_d2 = green_info["turning_point_d2"]
    free_surface_filter = green_info["free_surface_filter"]
    gravity_fc = green_info["gravity_fc"]
    gravity_harmonic = green_info["gravity_harmonic"]
    cal_sph = green_info["cal_sph"]
    cal_tor = green_info["cal_tor"]
    min_harmonic = green_info["min_harmonic"]
    max_harmonic = green_info["max_harmonic"]
    source_radius = green_info["source_radius"]
    source_duration = green_info["source_duration"]
    output_observables = green_info["output_observables"]
    time_window = green_info["time_window"]
    time_reduction = green_info["time_reduction"]
    path_nd = green_info["path_nd"]
    earth_model_layer_num = green_info["earth_model_layer_num"]
    physical_dispersion = green_info["physical_dispersion"]

    path_spec = str(
        os.path.join(
            path_green,
            "GreenSpec",
            "%.2f" % event_depth,
            "%.2f" % receiver_depth,
            "",
        )
    )
    if read_name == "hash":
        read_name = hash_read_pars(
            event_lat,
            event_lon,
            event_depth,
            receiver_lat,
            receiver_lon,
            receiver_depth,
        )
    path_func = str(os.path.join(path_green, "read", read_name, ""))
    os.makedirs(path_func, exist_ok=True)

    lines = str_inp.split("\n")
    lines = [line + "\n" for line in lines]
    last_line = [lines[-1]]
    lines_earth_head = lines[141:155]
    lines_earth = lines[155:-1]
    lines = lines[:140]  # cutoff locs

    lines[24] = "%.2f\n" % receiver_depth

    lines[52] = "%.2f  %.2f\n" % (spec_time_window, sampling_interval)
    lines[53] = "%.2f\n" % max_frequency
    lines[54] = "%.2f\n" % max_slowness
    lines[55] = "%.2f\n" % anti_alias
    lines[56] = "%d %.2f %.2f\n" % (
        turning_point_filter,
        turning_point_d1,
        turning_point_d2,
    )
    lines[57] = "6371.0 %d\n" % free_surface_filter

    lines[65] = "%f %d\n" % (gravity_fc, gravity_harmonic)

    lines[75] = "%d %d %d %d\n" % (cal_sph, cal_tor, min_harmonic, max_harmonic)

    lines[86] = "1 %f '%s'\n" % (source_radius, path_spec)
    lines[87] = "%.2f 'Green_%.2fkm' %d\n" % (event_depth, event_depth, 0)

    lines[111] = "1 1\n"
    mt_ned = check_convert_fm(focal_mechanism)
    mt = convert_mt_axis(mt=mt_ned, convert_flag="ned2rtp")
    lines[112] = (
        "1.0 "
        + " ".join("%f" % mt[_] for _ in range(6))
        + " 0.0 0.0 %.2f 0.0 %f\n" % (event_depth, source_duration)
    )
    lines[135] = " ".join(["%d" % output_observables[_] for _ in range(11)]) + "\n"
    lines[136] = "'%s'\n" % path_func
    lines[137] = "%f\n" % time_window
    lines[138] = "0 0 0\n"
    lines[139] = "0 %f\n" % max_slowness

    lines_receiver = [
        "1\n",
        "%.4f %.4f 'read' %.2f\n" % (receiver_lat, receiver_lon, time_reduction),
    ]

    lines = lines + lines_receiver
    path_inp = os.path.join(path_func, "read.inp")

    if path_nd is not None:
        lines_earth = convert_earth_model_nd2inp(
            path_nd=path_nd, path_output="earth_model.dat"
        )
    if earth_model_layer_num is None:
        earth_model_layer_num = len(lines_earth)
    lines_earth = lines_earth[:earth_model_layer_num]
    lines_earth_head[7] = "%d  %d\n" % (earth_model_layer_num, physical_dispersion)

    with open(path_inp, "w") as fw:
        fw.writelines(lines + lines_earth_head + lines_earth + last_line)
    return path_inp


def call_qssp2020_read(path_green, path_inp, check_finished=False):
    os.chdir(path_green)
    sub_dir = os.path.dirname(path_inp)
    hash_v = os.path.basename(sub_dir)
    path_inp = os.path.join(".", "read", hash_v, "read.inp").replace("'", "")
    path_finished = os.path.join(sub_dir, ".finished")
    if check_finished and os.path.exists(path_finished):
        return None

    if platform.system() == "Windows":
        spgrn_process = subprocess.Popen(
            [os.path.join(path_green, "qssp2020.exe")],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )
        spgrn_process.communicate(str.encode(path_inp))
    else:
        spgrn_process = subprocess.Popen(
            [os.path.join(path_green, "qssp2020.bin")],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )
        spgrn_process.communicate(str.encode(path_inp))

    with open(path_finished, "w") as fw:
        fw.writelines([])


def read_by_qssp(
    path_green,
    event_lat,
    event_lon,
    event_depth_km,
    receiver_lat,
    receiver_lon,
    receiver_depth_km,
    focal_mechanism,
    srate,
    read_name="hash",
    output_type="disp",
    green_info=None,
    check_finished=False,
):
    """ """
    if green_info is None:
        with open(os.path.join(path_green, "green_lib_info.json"), "r") as fr:
            green_info = json.load(fr)
    srate_grn = 1 / green_info["sampling_interval"]
    sampling_num = (
        round(green_info["time_window"] / green_info["sampling_interval"]) + 1
    )
    grn_dep_list = green_info["event_depth_list"]
    grn_receiver_list = green_info["receiver_depth_list"]
    if not isinstance(grn_dep_list, list):
        grn_dep_source = grn_dep_list
    else:
        grn_dep_source = grn_dep_list[
            np.argmin(np.abs(event_depth_km - np.array(grn_dep_list)))
        ]
    if not isinstance(grn_receiver_list, list):
        grn_dep_receiver = grn_receiver_list
    else:
        grn_dep_receiver = grn_receiver_list[
            np.argmin(np.abs(receiver_depth_km - np.array(grn_receiver_list)))
        ]

    path_inp = create_inp_qssp2020_read(
        read_name=read_name,
        path_green=path_green,
        event_lat=event_lat,
        event_lon=event_lon,
        event_depth=grn_dep_source,
        receiver_lat=receiver_lat,
        receiver_lon=receiver_lon,
        receiver_depth=grn_dep_receiver,
        focal_mechanism=focal_mechanism,
        green_info=green_info,
    )
    call_qssp2020_read(
        path_green=path_green, path_inp=path_inp, check_finished=check_finished
    )

    if output_type == "gravimeter":
        enz_list = [""]
    elif output_type in ["disp", "velo", "acce", "rota", "rota_rate", "gravitation"]:
        enz_list = ["_e", "_n", "_z"]
    elif output_type in ["stress", "stress_rate", "strain", "strain_rate"]:
        enz_list = ["_ee", "_en", "_ez", "_nn", "_nz", "_zz"]
    else:
        raise ValueError(
            "output_type must in  disp | velo | acce | strain | strain_rate | "
            "stress | stress_rate | rotation | rotation_rate | gravitation | gravimeter"
        )

    path_func = os.path.dirname(path_inp)
    data_enz = np.zeros((sampling_num, len(enz_list)))
    for i_enz in range(len(enz_list)):
        df = pd.read_csv(
            os.path.join(path_func, "_%s%s.dat" % (output_type, enz_list[i_enz])),
            sep="\\s+",
        )
        data_enz[:, i_enz] = df["read"].values

    seismograms = data_enz.T

    conv_shift = round(green_info["source_duration"] * srate_grn / 2)
    if conv_shift != 0:
        seismograms = np.roll(seismograms, -conv_shift)
        seismograms[:, -conv_shift:] = 0

    seismograms_resample = np.zeros(
        (seismograms.shape[0], round(sampling_num * srate / srate_grn))
    )
    for i in range(seismograms.shape[0]):
        seismograms_resample[i] = resample(
            seismograms[i], srate_old=srate_grn, srate_new=srate, zero_phase=True
        )

    return seismograms_resample



import os
import shutil
import platform
import pickle
import json
import datetime
from multiprocessing import Pool

import numpy as np
from tqdm import tqdm

try:
    from mpi4py import MPI
except:
    pass
import jpype

from .create_qseis06 import (
    create_dir_qseis06,
    create_inp_qseis06,
    create_inp_qseis06_points,
    call_qseis06,
    convert_pd2bin_qseis06,
)
from .pytaup import create_tpts_table
from .utils import group, convert_earth_model_nd2nd_without_Q


# 新增：imap_unordered 的打包调用助手
def _call_qseis06_star(args):
    return call_qseis06(*args)


def create_order_ind(order, diff_accu_order):
    if order == diff_accu_order // 2:
        order_ind = 0
    elif order < diff_accu_order // 2:
        order_ind = order + 1
    else:
        order_ind = order
    return order_ind


def pre_process_qseis06(
    processes_num,
    path_green,
    path_bin,
    event_depth_list,
    receiver_depth_list,
    dist_range,
    delta_dist,
    N_each_group,
    time_window,
    sampling_interval,
    slowness_int_algorithm=0,
    slowness_window=None,
    time_reduction_velo=0,
    wavenumber_sampling_rate=12,
    anti_alias=0.01,
    free_surface=True,
    wavelet_duration=0,
    wavelet_type=1,
    flat_earth_transform=True,
    path_nd=None,
    earth_model_layer_num=None,
    check_finished_tpts_table=False,
):
    print("preprocessing")
    os.makedirs(path_green, exist_ok=True)
    if platform.system() == "Windows":
        path_bin_call = os.path.join(path_green, "qseis06.exe")
    else:
        path_bin_call = os.path.join(path_green, "qseis06.bin")
    shutil.copy(path_bin, path_bin_call)

    N_dist, N_dist_group = None, None
    for event_depth in event_depth_list:
        for receiver_depth in receiver_depth_list:
            # print("creating green func dir, info, inp for event_depth=%.2f receiver_depth=%.2f" %
            #       (event_depth, receiver_depth))
            N_dist, N_dist_group = create_dir_qseis06(
                path_green,
                event_depth,
                receiver_depth,
                dist_range,
                delta_dist,
                N_each_group,
                0,
            )
            path_sub_dir = str(
                os.path.join(path_green, "%.2f" % event_depth, "%.2f" % receiver_depth)
            )
            create_inp_qseis06(
                path_sub_dir=path_sub_dir,
                event_depth=event_depth,
                receiver_depth=receiver_depth,
                dist_range=dist_range,
                delta_dist=delta_dist,
                N_dist=N_dist,
                N_dist_group=N_dist_group,
                N_each_group=N_each_group,
                time_window=time_window,
                sampling_interval=sampling_interval,
                slowness_int_algorithm=slowness_int_algorithm,
                slowness_window=slowness_window,
                time_reduction_velo=time_reduction_velo,
                wavenumber_sampling_rate=wavenumber_sampling_rate,
                anti_alias=anti_alias,
                free_surface=free_surface,
                wavelet_duration=wavelet_duration,
                wavelet_type=wavelet_type,
                flat_earth_transform=flat_earth_transform,
                path_nd=path_nd,
                earth_model_layer_num=earth_model_layer_num,
                order=0,
            )

    path_nd_without_Q = os.path.join(path_green, "noQ.nd")
    convert_earth_model_nd2nd_without_Q(path_nd, path_nd_without_Q)

    # creating tp and ts tables
    for event_depth in tqdm(event_depth_list, desc="Creating travel time tables"):
        for receiver_depth in receiver_depth_list:
            create_tpts_table(
                path_green,
                event_depth,
                receiver_depth,
                dist_range,
                delta_dist,
                path_nd_without_Q,
                check_finished_tpts_table,
            )
    if jpype.isJVMStarted():
        jpype.shutdownJVM()

    green_info = {
        "processes_num": processes_num,
        "event_depth_list": event_depth_list,
        "receiver_depth_list": receiver_depth_list,
        "dist_range": dist_range,
        "delta_dist": delta_dist,
        "N_dist": N_dist,
        "N_dist_group": N_dist_group,
        "N_each_group": N_each_group,
        "time_window": time_window,
        "sampling_interval": sampling_interval,
        "sampling_num": round(time_window / sampling_interval + 1),
        "slowness_int_algorithm": slowness_int_algorithm,
        "slowness_window": slowness_window,
        "time_reduction_velo": time_reduction_velo,
        "wavenumber_sampling_rate": wavenumber_sampling_rate,
        "anti_alias": anti_alias,
        "free_surface": free_surface,
        "wavelet_duration": wavelet_duration,
        "wavelet_type": wavelet_type,
        "flat_earth_transform": flat_earth_transform,
        "path_nd": path_nd,
        "path_nd_without_Q": path_nd_without_Q,
        "earth_model_layer_num": earth_model_layer_num,
    }
    json_str = json.dumps(green_info, indent=4, ensure_ascii=False)
    with open(
        os.path.join(path_green, "green_lib_info.json"), "w", encoding="utf-8"
    ) as file:
        file.write(json_str)

    inp_list = []
    for event_dep in event_depth_list:
        for receiver_dep in receiver_depth_list:
            for nn in range(N_dist_group):
                inp_list.append([event_dep, receiver_dep, nn, 0])
    group_list = group(inp_list, processes_num)
    with open(os.path.join(path_green, "group_list.pkl"), "wb") as fw:
        pickle.dump(group_list, fw)  # type: ignore
    return group_list


def pre_process_qseis06_strain_rate(
    processes_num,
    path_green,
    path_bin,
    event_depth_list,
    receiver_depth_list,
    dist_range,
    delta_dist,
    N_each_group,
    time_window,
    sampling_interval,
    slowness_int_algorithm=0,
    slowness_window=None,
    time_reduction_velo=0,
    wavenumber_sampling_rate=12,
    anti_alias=0.01,
    free_surface=True,
    wavelet_duration=0,
    wavelet_type=1,
    flat_earth_transform=True,
    path_nd=None,
    earth_model_layer_num=None,
    k_dr=0.001,
    dz=0.1,  # km
    diff_accu_order=4,
    check_finished_tpts_table=False,
):
    os.makedirs(path_green, exist_ok=True)
    if diff_accu_order not in [2, 4, 6, 8]:
        raise ValueError("diff_accu_order must be in [2,4,6,8]")
    if platform.system() == "Windows":
        path_bin_call = os.path.join(path_green, "qseis06.exe")
    else:
        path_bin_call = os.path.join(path_green, "qseis06.bin")
    shutil.copy(path_bin, path_bin_call)

    N_dist, N_dist_group = None, None
    for event_depth in event_depth_list:
        for receiver_depth in receiver_depth_list:
            print(
                "creating green func dir, info, inp for "
                "event_depth=%.2f receiver_depth=%.2f" % (event_depth, receiver_depth)
            )
            N_dist, N_dist_group = create_dir_qseis06(
                path_green,
                event_depth,
                receiver_depth,
                dist_range,
                delta_dist,
                N_each_group,
                diff_accu_order,
            )
            points = np.linspace(dist_range[0], dist_range[1], N_dist)
            for n_group in range(N_dist_group):
                for order in range(diff_accu_order + 1):
                    points_n_o = points[
                        n_group * N_each_group : (n_group + 1) * N_each_group
                    ]
                    points_n_o = (
                        points_n_o + (order - diff_accu_order // 2) * points_n_o * k_dr
                    )
                    order_ind = create_order_ind(order, diff_accu_order)
                    # print(n_group, order, order_ind, dist_range)
                    create_inp_qseis06_points(
                        path_green=path_green,
                        event_depth=event_depth,
                        receiver_depth=receiver_depth,
                        n_group=n_group,
                        points=points_n_o,
                        time_window=time_window,
                        sampling_interval=sampling_interval,
                        slowness_int_algorithm=slowness_int_algorithm,
                        slowness_window=slowness_window,
                        time_reduction_velo=time_reduction_velo,
                        wavenumber_sampling_rate=wavenumber_sampling_rate,
                        anti_alias=anti_alias,
                        free_surface=free_surface,
                        wavelet_duration=wavelet_duration,
                        wavelet_type=wavelet_type,
                        flat_earth_transform=flat_earth_transform,
                        path_nd=path_nd,
                        earth_model_layer_num=earth_model_layer_num,
                        order=order_ind,
                    )
                if receiver_depth > 0:
                    for order in range(diff_accu_order + 1):
                        if order == diff_accu_order // 2:
                            continue
                        receiver_depth_inp = (
                            receiver_depth + (order - diff_accu_order // 2) * dz
                        )
                        order_ind = (
                            create_order_ind(order, diff_accu_order) + diff_accu_order
                        )
                        path_sub_dir = str(
                            os.path.join(
                                path_green,
                                "%.2f" % event_depth,
                                "%.2f" % receiver_depth,
                            )
                        )
                        create_inp_qseis06(
                            path_sub_dir=path_sub_dir,
                            event_depth=event_depth,
                            receiver_depth=receiver_depth_inp,
                            dist_range=dist_range,
                            delta_dist=delta_dist,
                            N_dist=N_dist,
                            N_dist_group=N_dist_group,
                            N_each_group=N_each_group,
                            time_window=time_window,
                            sampling_interval=sampling_interval,
                            slowness_int_algorithm=slowness_int_algorithm,
                            slowness_window=slowness_window,
                            time_reduction_velo=time_reduction_velo,
                            wavenumber_sampling_rate=wavenumber_sampling_rate,
                            anti_alias=anti_alias,
                            free_surface=free_surface,
                            wavelet_duration=wavelet_duration,
                            wavelet_type=wavelet_type,
                            flat_earth_transform=flat_earth_transform,
                            path_nd=path_nd,
                            earth_model_layer_num=earth_model_layer_num,
                            order=order_ind,
                        )

    path_nd_without_Q = os.path.join(path_green, "noQ.nd")
    if path_nd is not None:
        convert_earth_model_nd2nd_without_Q(path_nd, path_nd_without_Q)

    # creating tp and ts tables
    for event_depth in tqdm(event_depth_list, desc="Creating travel time tables"):
        for receiver_depth in receiver_depth_list:
            create_tpts_table(
                path_green,
                event_depth,
                receiver_depth,
                dist_range,
                delta_dist,
                path_nd_without_Q,
                check_finished_tpts_table,
            )
    if jpype.isJVMStarted():
        jpype.shutdownJVM()

    green_info = {
        "processes_num": processes_num,
        "event_depth_list": event_depth_list,
        "receiver_depth_list": receiver_depth_list,
        "dist_range": dist_range,
        "delta_dist": delta_dist,
        "N_dist": N_dist,
        "N_dist_group": N_dist_group,
        "N_each_group": N_each_group,
        "time_window": time_window,
        "sampling_interval": sampling_interval,
        "sampling_num": round(time_window / sampling_interval + 1),
        "slowness_int_algorithm": slowness_int_algorithm,
        "slowness_window": slowness_window,
        "time_reduction_velo": time_reduction_velo,
        "wavenumber_sampling_rate": wavenumber_sampling_rate,
        "anti_alias": anti_alias,
        "free_surface": free_surface,
        "wavelet_duration": wavelet_duration,
        "wavelet_type": wavelet_type,
        "flat_earth_transform": flat_earth_transform,
        "path_nd": path_nd,
        "path_nd_without_Q": path_nd_without_Q,
        "earth_model_layer_num": earth_model_layer_num,
        "k_dr": k_dr,
        "dz": dz,
        "diff_accu_order": diff_accu_order,
    }
    json_str = json.dumps(green_info, indent=4, ensure_ascii=False)
    with open(
        os.path.join(path_green, "green_lib_info.json"), "w", encoding="utf-8"
    ) as file:
        file.write(json_str)

    inp_list = []
    for event_depth in event_depth_list:
        for receiver_depth in receiver_depth_list:
            for n_group in range(N_dist_group):
                if receiver_depth > 0:
                    for order in range(2 * diff_accu_order + 1):
                        inp_list.append([event_depth, receiver_depth, n_group, order])
                else:
                    for order in range(diff_accu_order + 1):
                        inp_list.append([event_depth, receiver_depth, n_group, order])
    inp_list_sorted = sorted(inp_list, key=lambda x: abs(x[0] - x[1]))
    group_list = group(inp_list_sorted, processes_num)
    with open(os.path.join(path_green, "group_list.pkl"), "wb") as fw:
        pickle.dump(group_list, fw)  # type: ignore
    return group_list


def create_grnlib_qseis06_sequential(
    path_green, check_finished=False, convert_pd2bin=True, remove_pd=True
):
    with open(os.path.join(path_green, "group_list.pkl"), "rb") as fr:
        group_list = pickle.load(fr)
    for item in tqdm(group_list, desc="Computing Green's Func Lib"):
        for i in range(len(item)):
            # print("computing " + str(item[i]))
            item[i] = item[i] + [path_green, check_finished]
            call_qseis06(*item[i])
    if convert_pd2bin:
        convert_pd2bin_qseis06_all(path_green, remove_pd)


def create_grnlib_qseis06_parallel(
    path_green, check_finished=False, convert_pd2bin=True, remove_pd=True
):
    with open(os.path.join(path_green, "group_list.pkl"), "rb") as fr:
        group_list = pickle.load(fr)
    # 展平任务
    tasks = []
    for grp in group_list:
        for item in grp:
            tasks.append(tuple(item + [path_green, check_finished]))

    # 尝试读取进程数
    processes = None
    try:
        with open(os.path.join(path_green, "green_lib_info.json"), "r") as fr:
            processes = json.load(fr).get("processes_num", None)
    except Exception:
        pass

    with Pool(processes=processes) as pool:
        for _ in tqdm(
            pool.imap_unordered(_call_qseis06_star, tasks, chunksize=1),
            total=len(tasks),
            desc="Computing Green's Func Lib",
        ):
            pass

    if convert_pd2bin:
        convert_pd2bin_qseis06_all(path_green, remove_pd)


def create_grnlib_qseis06_parallel_multi_nodes(path_green, check_finished=False):
    s = datetime.datetime.now()
    with open(os.path.join(path_green, "group_list.pkl"), "rb") as fr:
        group_list = pickle.load(fr)
    comm = MPI.COMM_WORLD
    processes_num = comm.Get_size()
    if processes_num != len(group_list[0]):
        raise ValueError(
            "processes_num is %d, item num in group is %d. \n"
            "Pleasse check the process num!" % (processes_num, len(group_list[0]))
        )
    rank = comm.Get_rank()
    ind_group = rank // processes_num
    ind_para = rank - ind_group * processes_num
    print(rank, ind_group, ind_para)
    call_qseis06(
        event_depth=group_list[ind_group][ind_para][0],
        receiver_depth=group_list[ind_group][ind_para][1],
        n_group=group_list[ind_group][ind_para][2],
        order=group_list[ind_group][ind_para][3],
        path_green=path_green,
        check_finished=check_finished,
    )
    e = datetime.datetime.now()
    print("run time:" + str(e - s))


def convert_pd2bin_qseis06_all(path_green, remove=False):
    print("converting ascii files to bytes files")
    with open(os.path.join(path_green, "green_lib_info.json"), "r") as fr:
        green_info = json.load(fr)
    event_depth_list = green_info["event_depth_list"]
    receiver_depth_list = green_info["receiver_depth_list"]
    for event_dep in event_depth_list:
        for receiver_dep in receiver_depth_list:
            sub_dir = str(
                os.path.join(path_green, "%.2f" % event_dep, "%.2f" % receiver_dep)
            )
            sub_sub_dirs = os.listdir(sub_dir)
            for sub_sub_dir in sub_sub_dirs:
                if "_table.bin" not in sub_sub_dir:
                    convert_pd2bin_qseis06(
                        os.path.join(sub_dir, sub_sub_dir), remove=remove
                    )


if __name__ == "__main__":
    pass



s = """#===============================================================================
# This is the input file of FORTRAN77 program "edcomp" for calculating
# earthquakes' static deformations (3 displacement components, 6 strain/stress
# components and 2 vertical tilt components) based on the dislocation theory.
# The earth model used is either a homogeneous or multi-layered, isotropic and
# elastic half-space. The earthquke source is represented by an arbitrary number
# of rectangular dislocation planes.
#
# Note that the Cartesian coordinate system is used and the seismological
# convention is adopted, that is, x is northward, y is eastward, and z is
# downward.
#
# First implemented in Potsdam, Feb, 1999
# Last modified: Potsdam, Nov, 2001
#
# by
# Rongjiang Wang, Frank Roth, & Francisco Lorenzo
# GeoForschungsZentrum Potsdam, Telegrafenberg, 14473 Potsdam, Germany
#
# For questions and suggestions please send e-mails to wang@gfz-potsdam.de
#===============================================================================
# OBSERVATION ARRAY
# =================
# 1. switch for irregular positions (0) or a 1D profile (1)
#    or a rectangular 2D observation array (2): ixyr
#
#    IF (the switch = 0 for irregular observation positions) THEN
#    
# 2. number of positions: nr
# 3. coordinates of the observations: (xr(i),yr(i)),i=1,nr
#
#    ELSE IF (the switch = 1 for a 1D profile) THEN
#
# 2. number of position samples of the profile: nr
# 3. the start and end positions: (xr1,yr1), (xr2,yr2)
#
#    ELSE IF (the switch = 2 for rectanglular 2D observation array) THEN
#
# 2. number of xr samples, start and end values [m]: nxr, xr1,xr2
# 3. number of yr samples, start and end values [m]: nyr, yr1,yr2
#
#    Note that the total number of observation positions (nr or nxr*nyr)
#    should be <= NRECMAX (see edcglobal.h)!
#===============================================================================
2
51  -35.00d+03   15.00d+03
51  -25.00d+03   25.00d+03
#===============================================================================
# OUTPUTS
# =======
# 1. output directory in char format: outdir
# 2. select the desired outputs (1/0 = yes/no)
# 3. the file names in char format for displacement vector, strain tensor,
#    stress tensor, vertical tilts, and los:
#    dispfile, strainfile, stressfile, tiltfile, losfile
#
#    Note that all file or directory names should not be longer than 80
#    characters. Directories must be ended by (unix) or (dos)!
#===============================================================================
''
0          0          1           0
'hs.disp'  'hs.strain'  'hs.stress'  'hs.tilt'
#===============================================================================
# RECTANGLAR DISLOCATION SOURCES
# ==============================
# 1. number of the source rectangles: ns (<= NSMAX in edcglobal.h)
# 2. the 6 parameters for the 1. source rectangle:
#    Slip [m],
#    coordinates of the upper reference point for strike (xs, ys, zs) [m],
#    length (strike direction) [m], and width (dip direction) [m],
#    strike [deg], dip [deg], and rake [deg];
# 3. ... for the 2. source ...
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#    Note that if one of the parameters length and width = 0, then a line source
#    will be considered and the dislocation parameter Slip has the unit m^2; if
#    both length and width = 0, then a point source will be considered and the
#    Slip has the unit m^3.
#===============================================================================
1
#         coord. origin:
#-------------------------------------------------------------------------------
# no  Slip   xs        ys       zs        length    width   strike   dip  rake
#-------------------------------------------------------------------------------
1  2.50 0.0d+00  0.0d+00  0.2d+03   11.d+03  10.0d+03   174.0  88.0  178.0
#===============================================================================
# If the earth model used is a layered half-space, then the numerical Green's
# function approach is applied. The Green's functions should have been prepared
# with the program "edgrn" before the program "edcmp" is started. In this case,
# the following input data give the addresses where the Green's functions have
# been stored and the grid side to be used for the automatic discretization
# of the finite rectangular sources.
#
# If the earth model used is a homogeneous half-space, then the analytical
# method of Okada (1992) is applied. In this case, the Green's functions are
# not needed, and the following input data give the shear modulus and the
# Poisson ratio of the model.
#===============================================================================
# CHOICE OF EARTH MODEL
# =====================
# 1. switch for layered (1) or homogeneous (0) model
#
#    IF (layered model) THEN
#
# 2. directory of the Green's functions and the three files for the
#    fundamental Green's functions: grndir, grnfiles(3);
#
#    Note that all file or directory names should not be longer than 80
#    characters. Directories must be ended by (unix) or (dos)!
#
#    ELSE (homogeneous model) THEN
#
# 2. the observation depth, the two Lame constants parameters of the homogeneous
#    model: zrec [m], lambda [Pa], mu [Pa]
#===============================================================================
#  1
#  ''  'edgrnhs.ss'  'edgrnhs.ds'  'edgrnhs.cl'
0
0.00d+00  30516224000  33701888000
#================================end of input==================================="""



import sys

import datetime

from pygrnwang.focal_mechanism import plane2mt
from pygrnwang.geo import d2km, cal_first_p_s
from pygrnwang.read_qssp2020 import seek_qssp2020
from pygrnwang.read_by_qssp import read_by_qssp

from scipy.io import savemat
from scipy import signal
import numpy as np

if __name__ == "__main__":
    import time

    path_green = "/e/grns_test/test_qssp_lzw"
    event_depth_km = 0
    receiver_depth_km = 0
    az_deg = 40
    dist_km = 200
    focal_mechanism = [0, 90, -180]
    source_array = np.array(
        [
            [
                0,
                0,
                event_depth_km,
                focal_mechanism[0],
                focal_mechanism[1],
                focal_mechanism[2],
                0.001,
                0.001,
                1,
                1,
            ]
        ]
    )
    srate = 1
    before_p = None
    pad_zeros = False
    shift = False
    rotate = True
    only_seismograms = False
    output_type = "strain"
    model_name = "/e/grns_test/test_qssp_lzw/noq.nd"
    # s = time.time()
    # for i in range(1):
    #     seismograms, tpts_table, first_p, first_s, grn_dep, grn_receiver, green_dist = (
    #         seek_qssp2020(
    #             path_green,
    #             event_depth_km,
    #             receiver_depth_km,
    #             az_deg,
    #             dist_km,
    #             focal_mechanism,
    #             srate,
    #             before_p,
    #             pad_zeros,
    #             shift,
    #             rotate,
    #             only_seismograms,
    #             output_type,
    #             model_name,
    #         )
    #     )
    # e = time.time()
    # print("run time:", e - s, "s")
    # #
    # # for i in range(seismograms.shape[0]):
    # #     w = signal.firwin(11, 0.1, fs=srate / 2)
    # #     seismograms[i] = signal.convolve(seismograms[i], w, "same")
    # #     # w = signal.windows.hann(srate)
    # #     # seismograms[i] = signal.fftconvolve(seismograms[i],w, 'same')
    # mu = 31126160000
    # area = 1  # 1e6
    # disp_source = 1
    # m0 = mu * area * disp_source
    # print(m0)
    # print(seismograms.shape)
    # print(seismograms[1, round(0.4 * dist_km * srate + 1)] * m0)
    # print(tpts_table, first_p, first_s, grn_dep, grn_receiver, green_dist)

    first_p, _ = cal_first_p_s(event_depth_km, dist_km, receiver_depth_km, model_name)

    s = datetime.datetime.now()
    seismograms_qssp = read_by_qssp(
        path_green=path_green,
        source_array=source_array,
        obs_point=[
            dist_km * np.cos(np.deg2rad(az_deg)) / d2km,
            dist_km * np.sin(np.deg2rad(az_deg)) / d2km,
            receiver_depth_km,
        ],
        srate=srate,
        read_name="hash",
        output_type=output_type,
        green_info=None,
        check_finished=False,
    )
    e = datetime.datetime.now()
    print(e - s)
    # print(grn_dep, grn_receiver)

    import matplotlib.pyplot as plt
    import matplotlib


    ind_com = 4
    plt.figure()
    t = np.linspace(
        0, len(seismograms_qssp[ind_com]) / srate, len(seismograms_qssp[ind_com]), endpoint=False
    )
    plt.plot(
        np.array([first_p, first_p]),
        np.array([np.min(seismograms_qssp[ind_com]), np.max(seismograms_qssp[ind_com])]),
        color="red",
    )
    # plt.plot(t, seismograms[ind_com], color="blue")
    plt.plot(t, seismograms_qssp[ind_com], color="green")
    #plt.legend(["read_by_python", "read_by_qssp"])
    plt.show()



import os
import glob
import pickle
import shutil
import platform
import json
import datetime
from multiprocessing import Pool

import numpy as np
from tqdm import tqdm
import jpype

try:
    from mpi4py import MPI
except:
    pass

from .create_qssp2020 import (
    mt_com_list,
    create_inp_qssp2020,
    create_dir_qssp2020,
    call_qssp2020,
    convert_pd2bin_qssp2020,
)
from .utils import group, convert_earth_model_nd2nd_without_Q
from .pytaup import create_tpts_table


# 新增：imap_unordered 的打包调用助手
def _call_qssp2020_star(args):
    return call_qssp2020(*args)


def pre_process_spec(
        processes_num,
        path_green,
        path_bin,
        event_depth_list,
        receiver_depth_list,
        spec_time_window,
        sampling_interval,
        max_frequency,
        max_slowness,
        anti_alias,
        turning_point_filter,
        turning_point_d1,
        turning_point_d2,
        free_surface_filter,
        gravity_fc,
        gravity_harmonic,
        cal_sph,
        cal_tor,
        min_harmonic,
        max_harmonic,
        source_radius,
        source_duration,
        time_window,
        time_reduction,
        dist_range,
        delta_dist,
        path_nd=None,
        earth_model_layer_num=None,
        physical_dispersion=0,
):
    if platform.system() == "Windows":
        path_bin_call = os.path.join(path_green, "qssp2020.exe")
    else:
        path_bin_call = os.path.join(path_green, "qssp2020.bin")
    shutil.copy(path_bin, path_bin_call)

    item_list_spec = []
    for event_depth in event_depth_list:
        for receiver_depth in receiver_depth_list:
            create_dir_qssp2020(event_depth, receiver_depth, path_green)
            create_inp_qssp2020(
                "spec",
                path_green,
                event_depth,
                receiver_depth,
                spec_time_window,
                sampling_interval,
                max_frequency,
                max_slowness,
                anti_alias,
                turning_point_filter,
                turning_point_d1,
                turning_point_d2,
                free_surface_filter,
                gravity_fc,
                gravity_harmonic,
                cal_sph,
                cal_tor,
                min_harmonic,
                max_harmonic,
                source_radius,
                1,
                source_duration,
                [0 for _ in range(11)],
                time_window,
                time_reduction,
                dist_range,
                delta_dist,
                path_nd,
                earth_model_layer_num,
                physical_dispersion,
            )
            item_list_spec.append([event_depth, receiver_depth, "spec"])

    group_list_spec = group(item_list_spec, processes_num)
    with open(os.path.join(path_green, "group_list_spec.pkl"), "wb") as fw:
        pickle.dump(group_list_spec, fw)  # type: ignore
    return group_list_spec


def pre_process_func(
        processes_num,
        path_green,
        path_bin,
        event_depth_list,
        receiver_depth_list,
        spec_time_window,
        sampling_interval,
        max_frequency,
        max_slowness,
        anti_alias,
        turning_point_filter,
        turning_point_d1,
        turning_point_d2,
        free_surface_filter,
        gravity_fc,
        gravity_harmonic,
        cal_sph,
        cal_tor,
        min_harmonic,
        max_harmonic,
        source_radius,
        source_duration,
        output_observables: list,
        time_window,
        time_reduction,
        dist_range,
        delta_dist,
        path_nd=None,
        earth_model_layer_num=None,
        physical_dispersion=0,
):
    if platform.system() == "Windows":
        path_bin_call = os.path.join(path_green, "qssp2020.exe")
    else:
        path_bin_call = os.path.join(path_green, "qssp2020.bin")
    if not os.path.exists(path_bin_call):
        shutil.copy(path_bin, path_bin_call)
    item_list_func = []
    for event_depth in event_depth_list:
        for receiver_depth in receiver_depth_list:
            for mt_com in mt_com_list:
                create_inp_qssp2020(
                    mt_com,
                    path_green,
                    event_depth,
                    receiver_depth,
                    spec_time_window,
                    sampling_interval,
                    max_frequency,
                    max_slowness,
                    anti_alias,
                    turning_point_filter,
                    turning_point_d1,
                    turning_point_d2,
                    free_surface_filter,
                    gravity_fc,
                    gravity_harmonic,
                    cal_sph,
                    cal_tor,
                    min_harmonic,
                    max_harmonic,
                    source_radius,
                    0,
                    source_duration,
                    output_observables,
                    time_window,
                    time_reduction,
                    dist_range,
                    delta_dist,
                    path_nd,
                    earth_model_layer_num,
                    physical_dispersion,
                )
                item_list_func.append([event_depth, receiver_depth, mt_com])

    group_list_func = group(item_list_func, processes_num)
    with open(os.path.join(path_green, "group_list_func.pkl"), "wb") as fw:
        pickle.dump(group_list_func, fw)  # type: ignore
    return group_list_func


def pre_process_qssp2020(
        processes_num,
        path_green,
        path_bin,
        event_depth_list,
        receiver_depth_list,
        spec_time_window,
        sampling_interval,
        max_frequency,
        max_slowness,
        anti_alias,
        turning_point_filter,
        turning_point_d1,
        turning_point_d2,
        free_surface_filter,
        gravity_fc,
        gravity_harmonic,
        cal_sph,
        cal_tor,
        min_harmonic,
        max_harmonic,
        source_radius,
        source_duration,
        output_observables: list,
        time_window,
        time_reduction,
        dist_range,
        delta_dist,
        path_nd=None,
        earth_model_layer_num=None,
        physical_dispersion=0,
        check_finished_tpts_table=False,
):
    print("Preprocessing")
    pre_process_spec(
        processes_num,
        path_green,
        path_bin,
        event_depth_list,
        receiver_depth_list,
        spec_time_window,
        sampling_interval,
        max_frequency,
        max_slowness,
        anti_alias,
        turning_point_filter,
        turning_point_d1,
        turning_point_d2,
        free_surface_filter,
        gravity_fc,
        gravity_harmonic,
        cal_sph,
        cal_tor,
        min_harmonic,
        max_harmonic,
        source_radius,
        source_duration,
        time_window,
        time_reduction,
        dist_range,
        delta_dist,
        path_nd,
        earth_model_layer_num,
        physical_dispersion,
    )
    pre_process_func(
        processes_num,
        path_green,
        path_bin,
        event_depth_list,
        receiver_depth_list,
        spec_time_window,
        sampling_interval,
        max_frequency,
        max_slowness,
        anti_alias,
        turning_point_filter,
        turning_point_d1,
        turning_point_d2,
        free_surface_filter,
        gravity_fc,
        gravity_harmonic,
        cal_sph,
        cal_tor,
        min_harmonic,
        max_harmonic,
        source_radius,
        source_duration,
        output_observables,
        time_window,
        time_reduction,
        dist_range,
        delta_dist,
        path_nd,
        earth_model_layer_num,
        physical_dispersion,
    )

    path_nd_without_Q = os.path.join(path_green, "noQ.nd")
    convert_earth_model_nd2nd_without_Q(path_nd, path_nd_without_Q)

    # creating tp and ts tables
    for event_depth in tqdm(event_depth_list, desc="Creating travel time tables"):
        for receiver_depth in receiver_depth_list:
            create_tpts_table(
                os.path.join(path_green, "GreenFunc"),
                event_depth,
                receiver_depth,
                dist_range,
                delta_dist,
                path_nd_without_Q,
                check_finished_tpts_table,
            )
    if jpype.isJVMStarted():
        jpype.shutdownJVM()

    green_info = {
        "processes_num": processes_num,
        "event_depth_list": event_depth_list,
        "receiver_depth_list": receiver_depth_list,
        "spec_time_window": spec_time_window,
        "sampling_interval": sampling_interval,
        "max_frequency": max_frequency,
        "max_slowness": max_slowness,
        "anti_alias": anti_alias,
        "turning_point_filter": turning_point_filter,
        "turning_point_d1": turning_point_d1,
        "turning_point_d2": turning_point_d2,
        "free_surface_filter": free_surface_filter,
        "gravity_fc": gravity_fc,
        "gravity_harmonic": gravity_harmonic,
        "cal_sph": cal_sph,
        "cal_tor": cal_tor,
        "min_harmonic": min_harmonic,
        "max_harmonic": max_harmonic,
        "source_radius": source_radius,
        "source_duration": source_duration,
        "output_observables": output_observables,
        "time_window": time_window,
        "sampling_num": round(time_window / sampling_interval) + 1,
        "time_reduction": time_reduction,
        "dist_range": dist_range,
        "delta_dist": delta_dist,
        "path_nd": path_nd,
        "path_nd_without_Q": path_nd_without_Q,
        "earth_model_layer_num": earth_model_layer_num,
        "physical_dispersion": physical_dispersion,
    }
    json_str = json.dumps(green_info, indent=4, ensure_ascii=False)
    with open(
            os.path.join(path_green, "green_lib_info.json"), "w", encoding="utf-8"
    ) as file:
        file.write(json_str)


def create_grnlib_qssp2020_sequential(
        path_green, cal_spec=True, check_finished=False, convert_pd2bin=True, remove_pd=True
):
    if cal_spec:
        with open(os.path.join(path_green, "group_list_spec.pkl"), "rb") as fr:
            group_list_spec = pickle.load(fr)
        for item in tqdm(
                group_list_spec,
                desc="Compute the Green\'s function library in the transformed domain.",
        ):
            for i in range(len(item)):
                item[i] = item[i] + [path_green, check_finished]
                call_qssp2020(*item[i])

    with open(os.path.join(path_green, "group_list_func.pkl"), "rb") as fr:
        group_list_func = pickle.load(fr)
    for item in tqdm(
            group_list_func, desc="Compute the Green\'s function library in the time domain."
    ):
        for i in range(len(item)):
            item[i] = item[i] + [path_green, check_finished]
            call_qssp2020(*item[i])

    if convert_pd2bin:
        convert_pd2bin_qssp2020_all(path_green)
    if remove_pd:
        remove_dat_files(path_green)


def create_grnlib_qssp2020_parallel(
        path_green, cal_spec=True, check_finished=False, convert_pd2bin=True, remove_pd=True
):
    tasks = []

    if cal_spec:
        with open(os.path.join(path_green, "group_list_spec.pkl"), "rb") as fr:
            group_list_spec = pickle.load(fr)
        for grp in group_list_spec:
            for item in grp:
                tasks.append(tuple(item + [path_green, check_finished]))

    processes = None
    try:
        with open(os.path.join(path_green, "green_lib_info.json"), "r") as fr:
            processes = json.load(fr).get("processes_num", None)
    except Exception:
        pass

    with Pool(processes=processes) as pool:
        for _ in tqdm(
                pool.imap_unordered(_call_qssp2020_star, tasks, chunksize=1),
                total=len(tasks),
                desc="Compute QSSP2020 Green\'s library in the transformed domain.",
        ):
            pass

    tasks = []
    with open(os.path.join(path_green, "group_list_func.pkl"), "rb") as fr:
        group_list_func = pickle.load(fr)
    for grp in group_list_func:
        for item in grp:
            tasks.append(tuple(item + [path_green, check_finished]))

    with Pool(processes=processes) as pool:
        for _ in tqdm(
                pool.imap_unordered(_call_qssp2020_star, tasks, chunksize=1),
                total=len(tasks),
                desc="Compute QSSP2020 Green\'s function library in the time domain.",
        ):
            pass

    if convert_pd2bin:
        convert_pd2bin_qssp2020_all(path_green)
    if remove_pd:
        remove_dat_files(path_green)


def create_grnlib_qssp2020_spec_parallel_multi_nodes(path_green, check_finished=False):
    s = datetime.datetime.now()
    with open(os.path.join(path_green, "group_list_spec.pkl"), "rb") as fr:
        group_list_spec = pickle.load(fr)
    N_all = 0
    for ind_group in range(len(group_list_spec)):
        N_all = N_all + len(group_list_spec[ind_group])
    for ind_group in range(len(group_list_spec)):
        comm = MPI.COMM_WORLD
        processes_num = comm.Get_size()
        rank = comm.Get_rank()
        if processes_num < len(group_list_spec[0]):
            raise ValueError(
                "processes_num is %d, item num in group is %d. \n"
                "Pleasse check the process num!"
                % (processes_num, len(group_list_spec[0]))
            )
        print(
            "computing spec lib ind_group:%d rank:%d event_depth:%.2f receiver_depth:%.2f"
            % (
                ind_group,
                rank,
                group_list_spec[ind_group][rank][0],
                group_list_spec[ind_group][rank][1],
            )
        )
        if ind_group * len(group_list_spec[0]) + rank < N_all:
            call_qssp2020(
                event_depth=group_list_spec[ind_group][rank][0],
                receiver_depth=group_list_spec[ind_group][rank][1],
                mt_com=group_list_spec[ind_group][rank][2],
                path_green=path_green,
                check_finished=check_finished,
            )
    e = datetime.datetime.now()
    print("run time:" + str(e - s))


def create_grnlib_qssp2020_func_parallel_multi_nodes(path_green, check_finished=False):
    s = datetime.datetime.now()
    with open(os.path.join(path_green, "group_list_func.pkl"), "rb") as fr:
        group_list_spec = pickle.load(fr)
    N_all = 0
    for ind_group in range(len(group_list_spec)):
        N_all = N_all + len(group_list_spec[ind_group])
    for ind_group in range(len(group_list_spec)):
        comm = MPI.COMM_WORLD
        processes_num = comm.Get_size()
        rank = comm.Get_rank()
        if processes_num < len(group_list_spec[0]):
            raise ValueError(
                "processes_num is %d, item num in group is %d. \n"
                "Pleasse check the process num!"
                % (processes_num, len(group_list_spec[0]))
            )
        print(
            "computing time lib ind_group:%d rank:%d event_depth:%.2f receiver_depth:%.2f mt_com:%s"
            % (
                ind_group,
                rank,
                group_list_spec[ind_group][rank][0],
                group_list_spec[ind_group][rank][1],
                group_list_spec[ind_group][rank][2],
            )
        )
        if ind_group * len(group_list_spec[0]) + rank < N_all:
            call_qssp2020(
                event_depth=group_list_spec[ind_group][rank][0],
                receiver_depth=group_list_spec[ind_group][rank][1],
                mt_com=group_list_spec[ind_group][rank][2],
                path_green=path_green,
                check_finished=check_finished,
            )
    e = datetime.datetime.now()
    print("run time:" + str(e - s))


def convert_pd2bin_qssp2020_all(path_green):
    print("converting ascii files to byte files")
    with open(os.path.join(path_green, "green_lib_info.json"), "r") as fr:
        green_info = json.load(fr)
    event_dep_list = green_info["event_depth_list"]
    receiver_dep_list = green_info["receiver_depth_list"]
    output_observables = np.nonzero(np.array(green_info["output_observables"]))[0]
    for i in range(len(event_dep_list)):
        for j in range(len(receiver_dep_list)):
            for output_type_ind in output_observables:
                convert_pd2bin_qssp2020(
                    path_green,
                    event_dep_list[i],
                    receiver_dep_list[j],
                    int(output_type_ind),
                )


def remove_dat_files(path_green):
    print("removing dat files")
    path_func = os.path.join(path_green, "GreenFunc")
    for root, dirs, files in os.walk(path_func):
        for file in glob.glob(os.path.join(root, "*.dat")):
            os.remove(file)


if __name__ == "__main__":
    pass



import os.path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from scipy.signal import firwin
from pygrnwang.signal_process import filter_butter
from pygrnwang.read_qseis2025 import *

if __name__ == "__main__":
    path_green = "/e/grns_test/test_qseis_vertical/"
    event_depth_km = 10
    receiver_depth_km = 1
    az_deg = 0
    dist_km = 0
    fm = [30, 40, 50]
    srate = 2
    cut_length = 256

    cutoff = 0.5
    numtaps_qseis = 21
    taps = firwin(numtaps_qseis, cutoff, fs=srate, window="hamming")

    (
        stress_rate_qseis,
        tpts_table,
        first_p,
        first_s,
        grn_dep_source,
        grn_dep_receiver,
        grn_dist,
    ) = seek_qseis2025(
        path_green=path_green,
        event_depth_km=event_depth_km,
        receiver_depth_km=receiver_depth_km,
        az_deg=az_deg,
        dist_km=dist_km,
        focal_mechanism=fm,
        srate=srate,
        before_p=None,
        pad_zeros=False,
        shift=False,
        rotate=False,
        only_seismograms=False,
        model_name=os.path.join(path_green, 'noq.nd'),
        green_info=None,
    )

    qseis_list = []
    qssp_list = []
    name_list = ["ee", "en", "ez", "nn", "nz", "zz"]

    for ind in range(6):
        # qseis_ind = resample(stress_rate_qseis[ind], srate_old=4, srate_new=srate)[:cut_length]
        qseis_ind = stress_rate_qseis[ind][:cut_length]
        # qseis_ind = np.convolve(qseis_ind, taps)[:cut_length]
        # qseis_ind = filter_butter(qseis_ind, srate, [0, cutoff])
        qseis_list.append(qseis_ind)

        df_qssp = pd.read_csv(
            '/e/grns_test/test_qssp_vertical/_stress_%s.dat' % name_list[ind], sep='\\s+')
        qssp_ind = df_qssp['test1'].values[:cut_length]
        qssp_list.append(qssp_ind)

    qseis_list = np.array(qseis_list)
    qssp_list = np.array(qssp_list)

    t = np.linspace(0, cut_length / srate, cut_length, endpoint=False)
    fig, axs = plt.subplots(6, 1)
    for ind in range(6):
        axs[ind].plot(t, qseis_list[ind], label="qseis", color="blue")
        axs[ind].plot(t, qssp_list[ind], label="qssp", color="green")
        axs[ind].set_ylabel("s%s" % name_list[ind])
    axs[-1].set_xlabel("Time (s)")
    axs[-1].legend()
    plt.show()



import numpy as np
from scipy import signal


def taper(data, taper_length=None, max_percentage=0.05) -> np.ndarray:
    data = data.copy()
    if taper_length is None:
        taper_length = max(2, round(len(data) * max_percentage))
    taper_window = signal.windows.hann(2 * taper_length)
    data[:taper_length] = data[:taper_length] * taper_window[:taper_length]
    data[-taper_length:] = data[-taper_length:] * taper_window[-taper_length:]
    return data


def cal_sos(srate, freq_band, butter_order=4):
    fn = srate / 2
    if (freq_band[0] == 0) and (freq_band[1] != 0) and (freq_band[1] / fn < 1):
        sos = signal.butter(
            butter_order, freq_band[1] / fn, btype="lowpass", output="sos"
        )
    elif (freq_band[0] != 0) and ((freq_band[1] == 0) or (freq_band[1] / fn >= 1)):
        sos = signal.butter(
            butter_order, freq_band[0] / fn, btype="highpass", output="sos"
        )
    elif (freq_band[0] != 0) and (freq_band[1] != 0) and (freq_band[1] / fn < 1):
        sos = signal.butter(
            butter_order,
            [freq_band[0] / fn, freq_band[1] / fn],
            btype="bandpass",
            output="sos",
        )
    else:
        sos = None
    return sos


def filter_butter(data: np.ndarray, srate, freq_band, butter_order=4, zero_phase=False):
    data = data.copy()
    sos = cal_sos(srate, freq_band, butter_order)
    if sos is not None:
        if zero_phase:
            data = signal.sosfiltfilt(sos, data)
        else:
            data = signal.sosfilt(sos, data)
    return data


def resample(data, srate_old: float, srate_new: float, zero_phase=True):
    if zero_phase:
        if float(srate_old).is_integer() and float(srate_new).is_integer():
            srate_old = int(srate_old)
            srate_new = int(srate_new)
            gcd = np.gcd(srate_new, srate_old)
            p = srate_new // gcd
            q = srate_old // gcd
            data = signal.resample_poly(data, p, q)
        else:
            data = signal.resample(data, round(len(data) * srate_new / srate_old))
            data = filter_butter(
                data=data,
                srate=srate_new,
                freq_band=[0, srate_old / 2],
                zero_phase=zero_phase,
            )
    else:
        q = srate_old / srate_new
        if srate_new < srate_old:
            if q.is_integer():
                data = signal.decimate(
                    data, q=int(q), ftype="fir", zero_phase=zero_phase
                )
            else:
                data = filter_butter(
                    data=data,
                    srate=srate_old,
                    freq_band=[0, srate_new / 2],
                    zero_phase=zero_phase,
                )
                data = signal.resample(x=data, num=round(len(data) * q))
        elif srate_new > srate_old:
            data = signal.resample(data, round(len(data) * q))
            data = filter_butter(
                data=data,
                srate=srate_new,
                freq_band=[0, srate_old / 2],
                zero_phase=zero_phase,
            )
    return data


def linear_interp(data, N_new) -> np.ndarray:
    points_loc = np.arange(0, len(data))
    points_loc_new = np.linspace(0, len(data), N_new, endpoint=False)
    data_new = np.interp(points_loc_new, points_loc, data)
    return data_new



import os
import subprocess
import platform

from .spgrn2020inp import s as str_inp
from .utils import convert_earth_model_nd2inp


def create_dir_spgrn2020(event_depth, receiver_depth, path_green):
    # 新建文件夹并返回路径,若文件夹已存在则直接返回路径
    os.makedirs(os.path.join(path_green, "GreenFunc"), exist_ok=True)
    os.makedirs(os.path.join(path_green, "GreenSpec"), exist_ok=True)
    path_func = os.path.join(
        path_green, "GreenFunc", "%.2f" % event_depth, "%.2f" % receiver_depth, ""
    )
    path_spec = os.path.join(
        path_green, "GreenSpec", "%.2f" % event_depth, "%.2f" % receiver_depth, ""
    )
    os.makedirs(path_func, exist_ok=True)
    os.makedirs(path_spec, exist_ok=True)
    return path_func, path_spec


def create_inp_spgrn2020(
    path_green,
    event_depth,
    receiver_depth,
    spec_time_window,
    sampling_interval,
    max_frequency,
    max_slowness,
    anti_alias,
    gravity_fc,
    gravity_harmonic,
    cal_sph,
    cal_tor,
    source_radius,
    cal_gf,
    time_window,
    green_before_p,
    source_duration,
    dist_range,
    delta_dist_range,
    path_nd=None,
    earth_model_layer_num=None,
    physical_dispersion=0,
):
    path_func = str(
        os.path.join(
            path_green, "GreenFunc", "%.2f" % event_depth, "%.2f" % receiver_depth, ""
        )
    )
    path_spec = str(
        os.path.join(
            path_green, "GreenSpec", "%.2f" % event_depth, "%.2f" % receiver_depth, ""
        )
    )

    lines = str_inp.split("\n")
    lines = [line + "\n" for line in lines]
    last_line = [lines[-1]]
    lines_earth = lines[113:-1]
    lines = lines[:113]  # cutoff earth model
    lines[25] = "%.2f\n" % receiver_depth
    lines[41] = "%f  %f\n" % (spec_time_window, sampling_interval)
    lines[42] = "%f\n" % max_frequency
    lines[43] = "%f\n" % max_slowness
    lines[44] = "%f\n" % anti_alias
    lines[52] = "%f %d\n" % (gravity_fc, gravity_harmonic)
    lines[60] = "%d %d\n" % (cal_sph, cal_tor)
    lines[71] = '"%s"\n' % path_spec
    lines[73] = '%.2f  %.2f  "grn_d%.2f"  %d\n' % (
        event_depth,
        source_radius,
        event_depth,
        cal_gf,
    )
    lines[90] = '"%s"\n' % path_func
    lines[91] = '"GreenInfo%.2f.dat"\n' % event_depth
    lines[94] = "%f  %f\n" % (time_window, sampling_interval)
    lines[95] = "%f\n" % -green_before_p
    lines[96] = "%f\n" % source_duration
    lines[98] = "%f  %f  %f  %f\n" % (
        dist_range[0],
        dist_range[1],
        delta_dist_range[0],
        delta_dist_range[1],
    )
    if path_nd is not None:
        lines_earth = convert_earth_model_nd2inp(
            path_nd=path_nd, path_output="earth_model.dat"
        )
    if earth_model_layer_num is None:
        earth_model_layer_num = len(lines_earth)
    lines[106] = "%d  %d\n" % (earth_model_layer_num, physical_dispersion)
    path_inp = os.path.join(path_func, "grn.inp")
    with open(path_inp, "w") as fw:
        fw.writelines(lines + lines_earth + last_line)
    return path_inp


def call_spgrn2020(event_depth, receiver_depth, path_green, check_finished=False):
    # print(event_depth, receiver_depth, path_green, check_finished)
    # os.chdir(path_green)
    path_inp = str(
        os.path.join(
            path_green,
            "GreenFunc",
            "%.2f" % event_depth,
            "%.2f" % receiver_depth,
            "grn.inp",
        )
    )

    sub_dir = os.path.dirname(path_inp)
    path_finished = os.path.join(sub_dir, ".finished")
    if check_finished and os.path.exists(path_finished):
        return None

    if platform.system() == "Windows":
        spgrn_process = subprocess.Popen(
            [os.path.join(path_green, "spgrn2020.exe")],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )
        spgrn_process.communicate(str.encode(path_inp))
    else:
        spgrn_process = subprocess.Popen(
            [os.path.join(path_green, "spgrn2020.bin")],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )
        spgrn_process.communicate(str.encode(path_inp))

    with open(path_finished, "w") as fw:
        fw.writelines([])


if __name__ == "__main__":
    pass



from pygrnwang.create_qseis06_bulk import *

if __name__ == '__main__':
    wavelet_type = 1
    wavelet_duration = 5
    sampling_interval = 0.25
    time_window = (1024 - 1) * sampling_interval
    wavenumber_sampling_rate = 12
    path_green = "/e/grns_test/qseis06"
    pre_process_qseis06(
        processes_num=1,
        path_green=path_green,
        path_bin="/home/zjc/python_works/pygrnwang/pygrnwang/exec/qseis_stress.bin",
        # event_depth_list=list(np.linspace(2.5, 37.5, 8)),
        # receiver_depth_list=list(np.linspace(2.5, 37.5, 8)),
        # dist_range=[2, 800],
        # delta_dist=2,
        event_depth_list=[10],
        receiver_depth_list=[1],
        dist_range=[2, 100],
        delta_dist=2,
        N_each_group=300,
        time_window=time_window,
        sampling_interval=sampling_interval,
        slowness_int_algorithm=0,
        slowness_window=[0, 0, 0, 0],
        time_reduction_velo=0,
        wavenumber_sampling_rate=wavenumber_sampling_rate,
        anti_alias=0.01,
        free_surface=True,
        wavelet_duration=wavelet_duration,
        wavelet_type=wavelet_type,
        flat_earth_transform=False,
        path_nd='/home/zjc/articles/Myanmar/myanmar.nd',
        earth_model_layer_num=25,
        # path_nd="/e/grns_test/test_lzw_edcmp/test.nd",
        # earth_model_layer_num=3,
    )
    create_grnlib_qseis06_parallel(
        path_green=path_green, check_finished=False
    )



import os

import numpy as np
import jpype.imports  # use jpype to call java class

jar_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "exec",
    "TauP-2.6.1.jar",
)
if not jpype.isJVMStarted():
    jpype.startJVM(classpath=[jar_path])
from edu.sc.seis.TauP import TauP_Time  # type: ignore


def taup_time_java(
    event_depth_km, dist_km, phases_list, receiver_depth_km=0, model_name="ak135"
):
    ttobj = TauP_Time(model_name)
    ttobj.setSourceDepth(event_depth_km)
    ttobj.setReceiverDepth(receiver_depth_km)
    ttobj.setPhaseNames(phases_list)
    ttobj.calculate(dist_km / 111.19492664455874)

    N_arr = ttobj.getNumArrivals()
    results = {"phase": [], "puristphase": [], "time": [], "rayparameter": []}
    for i in range(N_arr):
        arr = ttobj.getArrival(i)
        results["phase"].append(str(arr.getName()))
        results["puristphase"].append(str(arr.getPuristName()))
        results["time"].append(float(arr.getTime()))
        results["rayparameter"].append(float(arr.getRayParam()))
    return results


def cal_first_p(event_depth_km, dist_km, receiver_depth_km=0, model_name="ak135"):
    if event_depth_km < receiver_depth_km:
        event_depth_km, receiver_depth_km = receiver_depth_km, event_depth_km
    phases_list_p = ["p", "P", "pP", "Pg", "Pn", "Pdiff", "PKP"]
    ttobj_p = TauP_Time(model_name)
    ttobj_p.setSourceDepth(event_depth_km)
    ttobj_p.setReceiverDepth(receiver_depth_km)
    ttobj_p.setPhaseNames(phases_list_p)
    ttobj_p.calculate(dist_km / 111.19492664455874)
    arr_p = ttobj_p.getArrival(0)
    first_p = arr_p.getTime()
    return first_p


def cal_first_p_s(event_depth_km, dist_km, receiver_depth_km=0, model_name="ak135"):
    if event_depth_km < receiver_depth_km:
        event_depth_km, receiver_depth_km = receiver_depth_km, event_depth_km
    phases_list_p = ["p", "P", "pP", "Pg", "Pn", "Pdiff", "PKP"]
    ttobj_p = TauP_Time(model_name)
    ttobj_p.setSourceDepth(event_depth_km)
    ttobj_p.setReceiverDepth(receiver_depth_km)
    ttobj_p.setPhaseNames(phases_list_p)
    ttobj_p.calculate(dist_km / 111.19492664455874)
    arr_p = ttobj_p.getArrival(0)
    first_p = arr_p.getTime()

    phases_list_s = ["s", "S", "sS", "pS", "Sg", "Sn", "Sdiff", "SKS"]
    ttobj_s = TauP_Time(model_name)
    ttobj_s.setSourceDepth(event_depth_km)
    ttobj_s.setReceiverDepth(receiver_depth_km)
    ttobj_s.setPhaseNames(phases_list_s)
    ttobj_s.calculate(dist_km / 111.19492664455874)
    arr_s = ttobj_s.getArrival(0)
    first_s = arr_s.getTime()

    return first_p, first_s


def create_tpts_table(
    path_green,
    event_depth_km,
    receiver_depth_km,
    dist_range,
    delta_dist,
    model_name="ak135",
    check_finished=False,
):
    path_tp_table = os.path.join(
        path_green,
        "%.2f" % event_depth_km,
        "%.2f" % receiver_depth_km,
        "tp_table.bin",
    )
    path_ts_table = os.path.join(
        path_green,
        "%.2f" % event_depth_km,
        "%.2f" % receiver_depth_km,
        "ts_table.bin",
    )
    if (
        check_finished
        and os.path.exists(path_tp_table)
        and os.path.exists(path_ts_table)
    ):
        return
    dist_kms = np.linspace(
        dist_range[0],
        dist_range[1],
        round(np.ceil((dist_range[1] - dist_range[0]) / delta_dist)) + 1,
    )
    tp_table = np.zeros(len(dist_kms), dtype=np.float32)
    ts_table = np.zeros(len(dist_kms), dtype=np.float32)
    for i in range(len(dist_kms)):
        first_p, first_s = cal_first_p_s(
            event_depth_km, dist_kms[i], receiver_depth_km, model_name
        )
        tp_table[i] = first_p
        ts_table[i] = first_s
    tp_table.tofile(path_tp_table)
    ts_table.tofile(path_ts_table)


def read_tpts_table(path_green, event_depth_km, receiver_depth_km, ind):
    fr_tp = open(
        os.path.join(
            path_green,
            "%.2f" % event_depth_km,
            "%.2f" % receiver_depth_km,
            "tp_table.bin",
        ),
        "rb",
    )
    tp = np.fromfile(file=fr_tp, dtype=np.float32, count=1, offset=ind * 4)[0]
    fr_tp.close()

    fr_ts = open(
        os.path.join(
            path_green,
            "%.2f" % event_depth_km,
            "%.2f" % receiver_depth_km,
            "ts_table.bin",
        ),
        "rb",
    )
    ts = np.fromfile(file=fr_ts, dtype=np.float32, count=1, offset=ind * 4)[0]
    fr_ts.close()
    return float(tp), float(ts)


if __name__ == "__main__":
    pass



import os
import pickle
import shutil
import platform
import json
import datetime
from tqdm import tqdm  # 新增

try:
    from mpi4py import MPI
except:
    pass
from multiprocessing import Pool
from .create_spgrn import (
    create_dir_spgrn2020,
    create_inp_spgrn2020,
    call_spgrn2020,
)
from .read_green_info_spgrn2020 import read_green_info_spgrn2020
from .utils import group, convert_earth_model_nd2nd_without_Q


# 新增：imap_unordered 的打包调用助手
def _call_spgrn2020_star(args):
    return call_spgrn2020(*args)


def pre_process_spgrn2020(
    processes_num,
    path_green,
    path_bin,
    event_depth_list,
    receiver_depth_list,
    spec_time_window,
    sampling_interval,
    max_frequency,
    max_slowness,
    anti_alias,
    gravity_fc,
    gravity_harmonic,
    cal_sph,
    cal_tor,
    source_radius,
    cal_gf,
    time_window,
    green_before_p,
    source_duration,
    dist_range,
    delta_dist_range,
    path_nd=None,
    earth_model_layer_num=None,
    physical_dispersion=0,
):
    if platform.system() == "Windows":
        path_bin_call = os.path.join(path_green, "spgrn2020.exe")
    else:
        path_bin_call = os.path.join(path_green, "spgrn2020.bin")
    if not os.path.exists(path_bin_call):
        shutil.copy(path_bin, path_bin_call)

    item_list = []
    for event_depth in event_depth_list:
        for receiver_depth in receiver_depth_list:
            create_dir_spgrn2020(event_depth, receiver_depth, path_green)
            create_inp_spgrn2020(
                path_green,
                event_depth,
                receiver_depth,
                spec_time_window,
                sampling_interval,
                max_frequency,
                max_slowness,
                anti_alias,
                gravity_fc,
                gravity_harmonic,
                cal_sph,
                cal_tor,
                source_radius,
                cal_gf,
                time_window,
                green_before_p,
                source_duration,
                dist_range,
                delta_dist_range,
                path_nd,
                earth_model_layer_num,
                physical_dispersion,
            )
            item_list.append([event_depth, receiver_depth])
    group_list = group(item_list, processes_num)
    with open(os.path.join(path_green, "group_list.pkl"), "wb") as fw:
        pickle.dump(group_list, fw)  # type: ignore

    path_nd_without_Q = os.path.join(path_green, "noQ.nd")
    convert_earth_model_nd2nd_without_Q(path_nd, path_nd_without_Q)

    params = {
        "path_green": path_green,
        "path_bin": path_bin,
        "event_depth_list": event_depth_list,
        "receiver_depth_list": receiver_depth_list,
        "spec_time_window": spec_time_window,
        "sampling_interval": sampling_interval,
        "max_frequency": max_frequency,
        "max_slowness": max_slowness,
        "anti_alias": anti_alias,
        "gravity_fc": gravity_fc,
        "gravity_harmonic": gravity_harmonic,
        "cal_sph": cal_sph,
        "cal_tor": cal_tor,
        "source_radius": source_radius,
        "cal_gf": cal_gf,
        "time_window": time_window,
        "sampling_num": round(time_window / sampling_interval + 1),
        "green_before_p": green_before_p,
        "source_duration": source_duration,
        "dist_range": dist_range,
        "delta_dist_range": delta_dist_range,
        "path_nd": path_nd,
        "earth_model_layer_num": earth_model_layer_num,
        "physical_dispersion": physical_dispersion,
        "path_nd_without_Q": path_nd_without_Q,
    }
    json_str = json.dumps(params, indent=4, ensure_ascii=False)
    with open(
        os.path.join(path_green, "green_lib_info.json"), "w", encoding="utf-8"
    ) as file:
        file.write(json_str)

    return group_list


def update_green_info_lib_json(path_green, event_depth, receiver_depth):
    with open(os.path.join(path_green, "green_lib_info.json"), "r") as fr:
        green_info = json.load(fr)
    green_info_dep = read_green_info_spgrn2020(
        str(
            os.path.join(
                path_green, "GreenFunc", "%.2f" % event_depth, "%.2f" % receiver_depth
            )
        ),
        event_depth,
    )
    green_info["dist_list"] = green_info_dep["dist_list"]
    json_str = json.dumps(green_info, indent=4, ensure_ascii=False)
    with open(
        os.path.join(path_green, "green_lib_info.json"), "w", encoding="utf-8"
    ) as file:
        file.write(json_str)


def create_grnlib_spgrn2020_sequential(path_green, check_finished=False):
    s = datetime.datetime.now()
    with open(os.path.join(path_green, "group_list.pkl"), "rb") as fr:
        group_list = pickle.load(fr)
    for item in group_list:
        for i in range(len(item)):
            print("computing " + str(item[i]))
            call_spgrn2020(item[i][0], item[i][1], path_green, check_finished)
    update_green_info_lib_json(path_green, group_list[0][0][0], group_list[0][0][1])
    e = datetime.datetime.now()
    print("run time:%s" % str(e - s))


def create_grnlib_spgrn2020_parallel(path_green, check_finished=False):
    s = datetime.datetime.now()
    with open(os.path.join(path_green, "group_list.pkl"), "rb") as fr:
        group_list = pickle.load(fr)
    # 展平任务
    tasks = []
    for grp in group_list:
        for item in grp:
            tasks.append(tuple(item + [path_green, check_finished]))

    # 读取进程数（可选）
    processes = None
    try:
        with open(os.path.join(path_green, "green_lib_info.json"), "r") as fr:
            processes = json.load(fr).get("processes_num", None)
    except Exception:
        pass

    with Pool(processes=processes) as pool:
        for _ in tqdm(
            pool.imap_unordered(_call_spgrn2020_star, tasks, chunksize=1),
            total=len(tasks),
            desc="Computing SPGRN2020 library",
        ):
            pass

    update_green_info_lib_json(path_green, group_list[0][0][0], group_list[0][0][1])
    e = datetime.datetime.now()
    print("run time:" + str(e - s))


def create_grnlib_spgrn2020_parallel_multi_nodes(path_green, check_finished=False):
    s = datetime.datetime.now()
    with open(os.path.join(path_green, "group_list.pkl"), "rb") as fr:
        group_list = pickle.load(fr)
    for ind_group in range(len(group_list)):
        comm = MPI.COMM_WORLD
        processes_num = comm.Get_size()
        rank = comm.Get_rank()
        if processes_num != len(group_list[0]):
            raise ValueError(
                "processes_num is %d, item num in group is %d. \n"
                "Pleasse check the process num!" % (processes_num, len(group_list[0]))
            )
        print("ind_group:%d rank:%d" % (ind_group, rank))
        call_spgrn2020(
            event_depth=group_list[ind_group][rank][0],
            receiver_depth=group_list[ind_group][rank][1],
            path_green=path_green,
            check_finished=check_finished,
        )
    update_green_info_lib_json(path_green, group_list[0][0][0], group_list[0][0][1])
    e = datetime.datetime.now()
    print("run time:" + str(e - s))


if __name__ == "__main__":
    pass



import matplotlib.pyplot as plt
import numpy as np


def plot_seismograms(seismograms, srate, ylabel="u (m)", cut_length=None):
    if cut_length is None:
        length = len(seismograms[0])
    else:
        length = cut_length * srate
    t = np.arange(0, round(length / srate), 1 / srate)
    plt.figure()
    plt.subplot(3, 1, 1)
    plt.plot(t, seismograms[0][:length], color="red", linewidth=0.2)
    plt.subplot(3, 1, 2)
    plt.plot(t, seismograms[1][:length], color="green", linewidth=0.2)
    plt.subplot(3, 1, 3)
    plt.plot(t, seismograms[2][:length], color="blue", linewidth=0.2)
    plt.xlabel("t (s)")
    plt.ylabel(ylabel)
    plt.show()



import os
import json

import numpy as np

from .read_qseis06 import (
    read_time_series_qseis06_bin,
    read_time_series_qseis06_ascii,
    synthesize_qseis06,
)
from .create_qseis06_bulk import create_order_ind
from .focal_mechanism import check_convert_fm
from .geo import rotate_symmetric_tensor_series
from .pytaup import read_tpts_table
from .utils import read_nd, shift_green2real_tpts
from .signal_process import resample


def diff_central_1order(v_array, diff_accu_order):
    if diff_accu_order == 2:
        return (v_array[2] - v_array[0]) / 2
    elif diff_accu_order == 4:
        return (
            1 / 12 * v_array[0]
            - 2 / 3 * v_array[1]
            + 2 / 3 * v_array[3]
            - 1 / 12 * v_array[4]
        )
    elif diff_accu_order == 6:
        return (
            -1 / 60 * v_array[0]
            + 3 / 20 * v_array[1]
            - 3 / 4 * v_array[2]
            + 3 / 4 * v_array[4]
            - 3 / 20 * v_array[5]
            + 1 / 60 * v_array[6]
        )
    elif diff_accu_order == 8:
        return (
            1 / 280 * v_array[0]
            - 4 / 105 * v_array[1]
            + 1 / 5 * v_array[2]
            - 4 / 5 * v_array[3]
            + 4 / 5 * v_array[5]
            - 1 / 5 * v_array[6]
            + 4 / 105 * v_array[7]
            - 1 / 280 * v_array[8]
        )
    else:
        raise ValueError("diff_accu_order must be in [2,4,6,8]")


def seek_qseis06_strain_rate_diff(
    path_green,
    event_depth_km,
    receiver_depth_km,
    az_deg,
    dist_km,
    focal_mechanism,
    srate,
    before_p=None,
    pad_zeros=False,
    shift=False,
    only_seismograms=True,
    model_name="ak135fc",
    green_info=None,
):
    if green_info is None:
        with open(os.path.join(path_green, "green_lib_info.json"), "r") as fr:
            green_info = json.load(fr)
    srate_grn = 1 / green_info["sampling_interval"]
    sampling_num = green_info["sampling_num"]
    time_reduction_velo = green_info["time_reduction_velo"]
    dist_range = green_info["dist_range"]
    delta_dist = green_info["delta_dist"]
    num_each_group = green_info["N_each_group"]
    grn_dep_list = green_info["event_depth_list"]
    grn_receiver_list = green_info["receiver_depth_list"]
    diff_accu_order = green_info["diff_accu_order"]
    k_dr = green_info["k_dr"]
    dz = green_info["dz"] * 1e3
    if not isinstance(grn_dep_list, list):
        grn_dep_source = grn_dep_list
    else:
        grn_dep_source = grn_dep_list[
            np.argmin(np.abs(event_depth_km - np.array(grn_dep_list)))
        ]
    if not isinstance(grn_receiver_list, list):
        grn_dep_receiver = grn_receiver_list
    else:
        grn_dep_receiver = grn_receiver_list[
            np.argmin(np.abs(receiver_depth_km - np.array(grn_receiver_list)))
        ]

    path_greenfunc = str(
        os.path.join(path_green, "%.2f" % grn_dep_source, "%.2f" % grn_dep_receiver)
    )
    ind = round((dist_km - dist_range[0]) / delta_dist)
    ind_group = ind // num_each_group

    grn_dist = dist_range[0] + ind * delta_dist
    dr = grn_dist * k_dr * 1e3
    start_count = ind - ind_group * num_each_group

    [M11, M12, M13, M22, M23, M33] = check_convert_fm(focal_mechanism=focal_mechanism)

    exp = (M11 + M22 + M33) / 3
    ss1 = M12
    ss2 = (M11 - M22) / 2
    ds1 = M13
    ds2 = M23
    clvd = (-M11 / 2 - M22 / 2 + M33) / (3 / 2)

    az_rad = np.deg2rad(az_deg)
    sin_az, cos_az = np.sin(az_rad), np.cos(az_rad)
    sin_2az, cos_2az = np.sin(2 * az_rad), np.cos(2 * az_rad)

    m1 = [exp, ss1 * sin_2az + ss2 * cos_2az, ds1 * cos_az + ds2 * sin_az, clvd]
    m2 = [ss1 * cos_2az - ss2 * sin_2az, ds1 * sin_az - ds2 * cos_az]
    pm1_paz = [
        exp,
        ss1 * 2 * cos_2az + ss2 * (-2) * sin_2az,
        ds1 * (-1) * sin_az + ds2 * cos_az,
        clvd,
    ]
    pm2_paz = [
        ss1 * (-2) * sin_2az - ss2 * 2 * cos_2az,
        ds1 * cos_az - ds2 * (-1) * sin_az,
    ]

    # partial u_i / partial r
    velo_raw_dr = np.zeros((3 * (diff_accu_order + 1), sampling_num))
    for order in range(diff_accu_order + 1):
        order_ind = create_order_ind(order, diff_accu_order)
        path_greenfunc_order = str(
            os.path.join(path_greenfunc, "%d_%d" % (ind_group, order_ind))
        )
        path_bin = os.path.join(path_greenfunc_order, "grn_tz.npy")
        if os.path.exists(path_bin):
            time_series_list = read_time_series_qseis06_bin(
                path_greenfunc=path_greenfunc_order, start_count=start_count
            )
        else:
            time_series_list = read_time_series_qseis06_ascii(
                path_greenfunc=path_greenfunc_order, start_count=start_count
            )
        velo_dr = synthesize_qseis06(time_series_list=time_series_list, m1=m1, m2=m2)
        velo_raw_dr[order * 3 : (order + 1) * 3, :] = velo_dr
    pur_pr = diff_central_1order(velo_raw_dr[0:-1:3, :], diff_accu_order) / dr
    put_pr = diff_central_1order(velo_raw_dr[1:-1:3, :], diff_accu_order) / dr
    puz_pr = (
        diff_central_1order(velo_raw_dr[2 : len(velo_raw_dr) : 3, :], diff_accu_order)
        / dr
    )

    # partial u_i / partial theta
    path_greenfunc_order = str(os.path.join(path_greenfunc, "%d_%d" % (ind_group, 0)))
    path_bin = os.path.join(path_greenfunc_order, "grn_tz.npy")
    if os.path.exists(path_bin):
        time_series_list = read_time_series_qseis06_bin(
            path_greenfunc=path_greenfunc_order, start_count=start_count
        )
    else:
        time_series_list = read_time_series_qseis06_ascii(
            path_greenfunc=path_greenfunc_order, start_count=start_count
        )
    velo_dt = synthesize_qseis06(
        time_series_list=time_series_list, m1=pm1_paz, m2=pm2_paz
    )
    pur_pt = velo_dt[0]
    put_pt = velo_dt[1]
    puz_pt = velo_dt[2]

    # calculate err, ert, ett
    ur0 = velo_raw_dr[diff_accu_order // 2, :]
    ut0 = velo_raw_dr[diff_accu_order // 2 + 1, :]
    r = grn_dist * 1e3  # m
    e_rr = pur_pr
    e_tt = (put_pt + ur0) / r
    e_rt = 1 / 2 * (pur_pt / r + put_pr - ut0 / r)

    # partial u_i / partial z
    # calculate erz, etz, ezz
    if receiver_depth_km > 0:
        velo_raw_dz = np.zeros((3 * (diff_accu_order + 1), sampling_num))
        for order in range(diff_accu_order + 1):
            if order == diff_accu_order // 2:
                continue
            order_ind = create_order_ind(order, diff_accu_order) + diff_accu_order
            path_greenfunc_order = str(
                os.path.join(path_greenfunc, "%d_%d" % (ind_group, order_ind))
            )
            path_bin = os.path.join(path_greenfunc_order, "grn_tz.npy")
            if os.path.exists(path_bin):
                time_series_list = read_time_series_qseis06_bin(
                    path_greenfunc=path_greenfunc_order, start_count=start_count
                )
            else:
                time_series_list = read_time_series_qseis06_ascii(
                    path_greenfunc=path_greenfunc_order, start_count=start_count
                )
            velo_dz = synthesize_qseis06(
                time_series_list=time_series_list, m1=m1, m2=m2
            )
            velo_raw_dz[order * 3 : (order + 1) * 3, :] = velo_dz
        pur_pz = diff_central_1order(velo_raw_dz[0:-1:3, :], diff_accu_order) / dz
        put_pz = diff_central_1order(velo_raw_dz[1:-1:3, :], diff_accu_order) / dz
        puz_pz = (
            diff_central_1order(
                velo_raw_dz[2 : len(velo_raw_dr) : 3, :], diff_accu_order
            )
            / dz
        )
        e_rz = 1 / 2 * (pur_pz + puz_pr)
        e_tz = 1 / 2 * (put_pz + puz_pt / r)
        e_zz = puz_pz
    else:
        path_nd_noQ = green_info["path_nd_without_Q"]
        nd_model = read_nd(path_nd_noQ)
        depth_array = nd_model[:, 0]
        ind = np.argmin(np.abs(depth_array - receiver_depth_km))
        vp = nd_model[ind, 1] * 1e3
        vs = nd_model[ind, 2] * 1e3
        rho = nd_model[ind, 3] * 1e3
        mu = rho * vs**2
        lam = rho * vp**2 - 2 * mu
        e_rz = np.zeros_like(e_rr)
        e_tz = np.zeros_like(e_rr)
        e_zz = -lam / (lam + 2 * mu) * (e_rr + e_tt)

    strain_tensor = np.array([e_tt, e_rt, -e_tz, e_rr, -e_rz, e_zz])
    seismograms = rotate_symmetric_tensor_series(strain_tensor.T, np.deg2rad(-az_deg)).T

    tpts_table = None
    if (before_p is not None) or shift or pad_zeros:
        first_p_grn, first_s_grn = read_tpts_table(
            path_green=path_green,
            event_depth_km=grn_dep_source,
            receiver_depth_km=grn_dep_receiver,
            ind=ind,
        )
        tpts_table = {"p_onset": first_p_grn, "s_onset": first_s_grn}

    if time_reduction_velo != 0:
        time_reduction = grn_dist / time_reduction_velo
    else:
        time_reduction = 0

    ts_count = 0
    if before_p is not None:
        ts_count = round(
            (tpts_table["p_onset"] - time_reduction - before_p) * srate_grn
        )
    if pad_zeros:
        if before_p is not None:
            raise ValueError("can not set before_p and pad_zeros together")
        ts_count = round(-time_reduction * srate_grn)
    seismograms = np.roll(seismograms, -ts_count)
    if ts_count > 0:
        seismograms[:, -ts_count:] = 0
    elif ts_count < 0:
        seismograms[:, :-ts_count] = 0

    first_p = None
    first_s = None
    if shift:
        seismograms, first_p, first_s = shift_green2real_tpts(
            seismograms=seismograms,
            tpts_table=tpts_table,
            srate=srate_grn,
            green_before_p=tpts_table["p_onset"] - time_reduction,
            event_depth_km=event_depth_km,
            dist_in_km=dist_km,
            receiver_depth_km=receiver_depth_km,
            model_name=model_name,
        )

    seismograms_resample = np.zeros((6, round(sampling_num * srate / srate_grn)))
    for i in range(6):
        seismograms_resample[i] = resample(
            seismograms[i], srate_old=srate_grn, srate_new=srate, zero_phase=True
        )

    if only_seismograms:
        return seismograms_resample
    else:
        return (
            seismograms_resample,
            tpts_table,
            first_p,
            first_s,
            grn_dep_source,
            grn_dep_receiver,
            grn_dist,
        )


def convert_strain2stress(strain, lam, mu):
    # ee en ez nn nz zz
    stress = np.zeros_like(strain)
    s = strain[0, :] + strain[3, :] + strain[5, :]
    stress[0, :] = lam * s + 2 * mu * strain[0, :]
    stress[1, :] = 2 * mu * strain[1, :]
    stress[2, :] = 2 * mu * strain[2, :]
    stress[3, :] = lam * s + 2 * mu * strain[3, :]
    stress[4, :] = 2 * mu * strain[4, :]
    stress[5, :] = lam * s + 2 * mu * strain[5, :]
    return stress


def seek_qseis06_stress_rate_diff(
    path_green,
    event_depth_km,
    receiver_depth_km,
    az_deg,
    dist_km,
    focal_mechanism,
    srate,
    before_p=None,
    pad_zeros=False,
    shift=False,
    only_seismograms=True,
    model_name="ak135fc",
    green_info=None,
):
    if green_info is None:
        with open(os.path.join(path_green, "green_lib_info.json"), "r") as fr:
            green_info = json.load(fr)
    srate_grn = 1 / green_info["sampling_interval"]
    sampling_num = (
        round(green_info["time_window"] / green_info["sampling_interval"]) + 1
    )
    (
        strain_rate,
        tpts_table,
        first_p,
        first_s,
        grn_dep_source,
        grn_dep_receiver,
        grn_dist,
    ) = seek_qseis06_strain_rate_diff(
        path_green,
        event_depth_km,
        receiver_depth_km,
        az_deg,
        dist_km,
        focal_mechanism,
        srate_grn,
        before_p,
        pad_zeros,
        shift,
        False,
        model_name,
        green_info,
    )

    path_nd_noQ = green_info["path_nd_without_Q"]
    nd_model = read_nd(path_nd_noQ)
    depth_array = nd_model[:, 0]
    ind = np.argmin(np.abs(depth_array - receiver_depth_km))
    vp = nd_model[ind, 1] * 1e3
    vs = nd_model[ind, 2] * 1e3
    rho = nd_model[ind, 3] * 1e3
    mu = rho * vs**2
    lam = rho * vp**2 - 2 * mu
    seismograms = convert_strain2stress(strain_rate, lam, mu)

    seismograms_resample = np.zeros((6, round(sampling_num * srate / srate_grn)))
    for i in range(6):
        seismograms_resample[i] = resample(
            seismograms[i], srate_old=srate_grn, srate_new=srate, zero_phase=True
        )

    if only_seismograms:
        return seismograms_resample
    else:
        return (
            seismograms_resample,
            tpts_table,
            first_p,
            first_s,
            grn_dep_source,
            grn_dep_receiver,
            grn_dist,
        )


if __name__ == "__main__":
    pass



import os
import math

from .edgrn2inp import s as str_inp
from .utils import convert_earth_model_nd2inp, call_exe


def create_inp_edgrn2(
    path_green,
    obs_depth,
    grn_dist_range,
    grn_delta_dist,
    grn_source_depth_range,
    grn_delta_source_depth,
    wavenumber_sampling_rate=12,
    path_nd=None,
    earth_model_layer_num=None,
):
    lines = str_inp.split("\n")
    lines = [line + "\n" for line in lines]
    lines_earth = lines[84:-1]
    lines_end = lines[-1]
    lines = lines[:84]

    lines[39] = "%.2f\n" % (obs_depth * 1e3)
    n_dist = math.ceil((grn_dist_range[1] - grn_dist_range[0]) / grn_delta_dist) + 1
    lines[40] = "%d %f %f\n" % (
        n_dist,
        grn_dist_range[0] * 1e3,
        grn_dist_range[1] * 1e3,
    )
    n_source_depth = (
        math.ceil(
            (grn_source_depth_range[1] - grn_source_depth_range[0])
            / grn_delta_source_depth
        )
        + 1
    )
    lines[41] = "%d %f %f\n" % (
        n_source_depth,
        grn_source_depth_range[0] * 1e3,
        grn_source_depth_range[1] * 1e3,
    )
    lines[51] = "%d\n" % wavenumber_sampling_rate
    path_edgrn_obs_dep = str(os.path.join(path_green, "edgrn2", "%.2f" % obs_depth, ""))
    lines[61] = "'%s' 'edgrn.ss' 'edgrn.ds' 'edgrn.cl'\n" % (path_edgrn_obs_dep)

    if path_nd is not None:
        lines_earth = convert_earth_model_nd2inp(
            path_nd=path_nd, path_output="earth_model.dat"
        )
    for i in range(len(lines_earth)):
        temp = lines_earth[i].split()
        lines_earth[i] = (
            temp[0] + " " + " ".join("%.4f" % (float(_) * 1e3) for _ in temp[1:-2])
        )
        lines_earth[i] = lines_earth[i] + "\n"
    if earth_model_layer_num is None:
        earth_model_layer_num = len(lines_earth)
    else:
        lines_earth = lines_earth[:earth_model_layer_num]
    lines[80] = "%d\n" % earth_model_layer_num

    lines = lines + lines_earth + [lines_end]
    with open(os.path.join(path_edgrn_obs_dep, "grn.inp"), "w") as fw:
        fw.writelines(lines)


def call_edgrn2(obs_depth, path_green, check_finished=False):
    os.chdir(path_green)
    sub_sub_dir = str(os.path.join("edgrn2", "%.2f" % obs_depth))
    if (
        check_finished
        and os.path.exists(os.path.join(sub_sub_dir, ".finished"))
        and len(os.listdir(sub_sub_dir)) > 2
    ):
        return None
    path_inp = str(os.path.join(sub_sub_dir, "grn.inp"))
    path_finished = os.path.join(sub_sub_dir, ".finished")

    if (
        check_finished
        and os.path.exists(path_finished)
        and len(os.listdir(sub_sub_dir)) > 2
    ):
        with open(path_finished, "r") as fr:
            output = fr.readlines()
        return output

    output = call_exe(
        path_green=path_green,
        path_inp=path_inp,
        path_finished=path_finished,
        name="edgrn2",
    )
    return output



import os.path
import sys

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from scipy.signal import firwin
from pygrnwang.signal_process import filter_butter, linear_interp
from pygrnwang.read_qseis2025 import *
from pygrnwang.geo import d2km
from pygrnwang.read_by_qssp import read_by_qssp
from pygrnwang.read_qssp2020 import seek_qssp2020
from pygrnwang.read_edcmp import *
from pygrnwang.create_edcmp_bulk import *
from pygrnwang.signal_process import *
from pygrnwang.obspy_taup import cal_first_p
from pygrnwang.focal_mechanism import *


def correct_zero_frequency(data, srate, A0, f_c, tc1, tc2,
                           ratio_interp=0):
    N_data = len(data)
    data = data[tc1:tc2]
    data[0] = 0
    data[-1] = 0
    # data = taper(data)
    if ratio_interp > 0:
        # u = np.concatenate([np.zeros(pad_len // 2), data.copy(), np.zeros(pad_len - pad_len // 2)])
        # l = ratio_interp * (tc2 - tc1)
        u = resample(data=data, srate_old=srate, srate_new=srate * ratio_interp)
        uf = np.fft.fft(u) / (srate * ratio_interp)
    else:
        u = data.copy()
        uf = np.fft.fft(u) / srate
    uf_correct = uf.copy()
    uf_correct[0] = A0

    N = len(uf)
    A_f = np.abs(uf)
    phi_f = np.angle(uf)

    # f = np.fft.fftfreq(N, 1 / srate)[:N // 2]
    # f_c = max(2, np.argmin(np.abs(cut_freq - f)))
    # print(f_c)

    w = np.zeros(N)
    w[0:f_c + 1] = 1 - 1 / 2 * (1 + np.cos(np.pi * np.arange(f_c + 1) / f_c))  # 0->1
    w[-f_c:] = w[1:f_c + 1][::-1]

    uf_correct[1:f_c + 1] = ((1 - w[1:f_c + 1]) * np.abs(A0) + w[1:f_c + 1] * A_f[1:f_c + 1]
                             ) * np.exp(1j * np.complex128(phi_f[1:f_c + 1]))
    uf_correct[-f_c:] = ((1 - w[-f_c:]) * np.abs(A0) + w[-f_c:] * A_f[-f_c:]
                         ) * np.exp(1j * np.complex128(phi_f[-f_c:]))

    if ratio_interp > 0:
        u_correct = np.real(np.fft.ifft(uf_correct)) * srate * ratio_interp
        # u_correct = filter_butter(data=u_correct, srate=srate * ratio_interp,
        #                           freq_band=[0, srate / 2])
        u_correct = resample(data=u_correct, srate_old=srate * ratio_interp, srate_new=srate)
    else:
        u_correct = np.real(np.fft.ifft(uf_correct)) * srate
    # u_correct = u_correct - u_correct[0]
    # u_correct = taper(u_correct, taper_len)
    u_correct = np.concatenate([np.zeros(tc1), u_correct, np.zeros(N_data - tc2)])
    return u_correct


if __name__ == "__main__":
    correct = True
    wavelet_type = 1
    wavelet_duration = 5
    sampling_interval = 0.25
    time_window_qseis = (1024 - 1) * sampling_interval
    wavenumber_sampling_rate = 12
    time_window_qssp = 256 - sampling_interval
    max_frequency = 2
    max_slowness = 0.4
    min_harmonic = 20000
    path_green_qseis = "/e/grns_test/qseis_%d_%d_%.3f_%.3f_%d_" % (
        wavelet_type, wavelet_duration, sampling_interval, time_window_qseis,
        wavenumber_sampling_rate
    )
    #path_green_qseis = "/e/grns_test/qseis_1_5_0.250_255.750_12_m"
    path_green_qssp = "/e/grns_test/qssp_%d_%.3f_%.3f_%.3f_%.3f_%d" % (
        wavelet_duration, sampling_interval, time_window_qssp,
        max_frequency, max_slowness, min_harmonic
    )
    path_green_edcmp = "/e/grns_test/test_edcmp"
    M0 = 31126160000 * (1 ** 2) * 1e6
    event_depth_km = 10
    receiver_depth_km = 1
    az_deg = 0
    dist_km = 6
    fm = [45, 90, 90]
    srate = 4
    cut_length = 128
    output_type = 'stress_rate'
    print(
        "%f %f"
        % (
            dist_km / d2km * np.cos(np.deg2rad(az_deg)),
            dist_km / d2km * np.sin(np.deg2rad(az_deg)),
        )
    )

    cutoff = 0.4
    numtaps = 51
    stf = firwin(numtaps, cutoff, fs=srate, window="hamming")
    stf = stf / (np.sum(stf) / srate)

    stress_qseis = seek_qseis2025(
        path_green=path_green_qseis,
        event_depth_km=event_depth_km,
        receiver_depth_km=receiver_depth_km,
        az_deg=az_deg,
        dist_km=dist_km,
        focal_mechanism=fm,
        srate=srate,
        output_type=output_type,
        before_p=None,
        pad_zeros=True,
        rotate=True,
        only_seismograms=True,
        model_name=os.path.join(path_green_qseis, "noq.nd"),
        green_info=None,
    )
    # stress_qseis = np.roll(stress_qseis, -40, axis=1)/10
    # fm = check_convert_fm(focal_mechanism=fm)
    # fm = mt2plane(fm)[0]
    source_array = np.array(
        [
            [
                0,
                0,
                event_depth_km,
                fm[0],
                fm[1],
                fm[2],
                0,
                0,
                1,
            ]
        ]
    )
    obs_point = np.array(
        [
            dist_km / d2km * np.cos(np.deg2rad(az_deg)),
            dist_km / d2km * np.sin(np.deg2rad(az_deg)),
            receiver_depth_km,
        ]
    )

    stress_qssp = seek_qssp2020(
        path_green=path_green_qssp,
        event_depth_km=event_depth_km,
        receiver_depth_km=receiver_depth_km,
        az_deg=az_deg,
        dist_km=dist_km,
        focal_mechanism=fm,
        srate=srate,
        before_p=None,
        pad_zeros=True,
        shift=False,
        rotate=True,
        only_seismograms=True,
        output_type=output_type,
        model_name=os.path.join(path_green_qssp, "noq.nd"),
        green_info=None
    )
    # stress_qssp_direct = read_by_qssp(
    #     path_green=path_green_qssp,
    #     event_lat=0,
    #     event_lon=0,
    #     event_depth_km=event_depth_km,
    #     receiver_lat=obs_point[0],
    #     receiver_lon=obs_point[1],
    #     receiver_depth_km=receiver_depth_km,
    #     focal_mechanism=fm,
    #     srate=srate,
    #     read_name="hash",
    #     output_type=output_type,
    #     green_info=None,
    #     check_finished=False
    # )

    path_bin_edcmp = "/home/zjc/python_works/pygrnwang/pygrnwang/exec/edcmp2.bin"
    source_array_new = np.zeros((len(source_array), 9))
    source_array_new[:, 0] = source_array[:, 8]
    source_array_new[:, 1:9] = source_array[:, 0:8]
    pre_process_edcmp2(
        processes_num=1,
        path_green=path_green_edcmp,
        path_bin=path_bin_edcmp,
        obs_depth_list=[receiver_depth_km],
        obs_x_range=[obs_point[0] - 0.01, obs_point[0] + 0.01],
        obs_y_range=[obs_point[1] - 0.01, obs_point[1] + 0.01],
        obs_delta_x=0.001,
        obs_delta_y=0.001,
        source_array=source_array_new,
        source_ref=[0, 0],
        obs_ref=[0, 0],
        layered=True,
    )
    compute_static_stress_edcmp2_sequential(
        path_green=path_green_edcmp, check_finished=False)
    stress_edcmp = seek_edcmp2(
        path_green_edcmp,
        'stress',
        np.array([obs_point]),
        geo_coordinate=True,
    )[0] * 1e6

    qseis_list = []
    qssp_list = []
    qssp_direct_list = []
    name_list = ["ee", "en", "ez", "nn", "nz", "zz"]
    first_p = cal_first_p(
        event_depth_km=event_depth_km,
        dist_km=dist_km,
        receiver_depth_km=receiver_depth_km,
        model_name='ak135'
    )
    tc1 = max(1, round(first_p * srate - 1))
    tc2 = round(dist_km * max_slowness * srate +
                1.5 * wavelet_duration * sampling_interval * srate+len(stf))
    print(tc1, tc2)
    for ind in range(6):
        qseis_ind = stress_qseis[ind][:cut_length] * M0
        #qseis_ind[:tc1] = 0
        #qseis_ind = np.convolve(qseis_ind, stf)[:cut_length] / srate
        if correct:
            qseis_ind = correct_zero_frequency(
                qseis_ind, srate, stress_edcmp[ind], 4,
                tc1, tc2, ratio_interp=0
            )

        # qseis_ind = np.convolve(qseis_ind, stf)[:cut_length]/srate
        # qseis_ind = filter_butter(qseis_ind, srate, [0, cutoff])
        # qseis_ind = np.cumsum(qseis_ind) / srate
        qseis_list.append(qseis_ind)

        qssp_ind = stress_qssp[ind][:cut_length] * M0
        qssp_ind = np.concatenate([np.zeros(1), qssp_ind])[:cut_length]
        #qssp_ind = np.convolve(qssp_ind, stf)[:cut_length]/srate
        # qssp_ind = filter_butter(qssp_ind, srate, [0, cutoff])
        # qssp_ind = np.cumsum(qssp_ind) / srate
        qssp_list.append(qssp_ind)

        # qssp_direct_ind = stress_qssp_direct[ind][round(20*srate):cut_length+round(20*srate)]*M0
        # #qssp_direct_ind = stress_qssp_direct[ind][:cut_length]
        # qssp_direct_ind = np.concatenate([np.zeros(1), qssp_direct_ind])[:cut_length]
        # qssp_direct_ind = np.convolve(qssp_direct_ind, stf)[:cut_length]
        # #qssp_direct_ind = filter_butter(qssp_direct_ind, srate, [0, cutoff])
        # # qssp_direct_ind = np.cumsum(qssp_direct_ind)/srate
        # qssp_direct_list.append(qssp_direct_ind)
    qseis_list = np.array(qseis_list)
    qssp_list = np.array(qssp_list)
    qssp_direct_list = np.array(qssp_direct_list)

    A_max = max(np.max(np.abs(qseis_list)), np.max(np.abs(qssp_list))) * 1.1
    t = np.linspace(0, cut_length / srate, cut_length, endpoint=False)
    fig, axs = plt.subplots(6, 1)
    for ind in range(6):
        axs[ind].plot(t, qseis_list[ind], label="qseis", color="red")
        axs[ind].plot(t, qssp_list[ind], label="qssp", color="green")
        # axs[ind].plot(t, qssp_direct_list[ind], label="qssp_direct", color="blue")
        # axs[ind].set_ylim([-A_max, A_max])
        axs[ind].set_ylabel("sr_%s" % name_list[ind])
    axs[-1].set_xlabel("Time (s)")
    axs[-1].legend(loc="upper right")
    plt.show()

    f = np.fft.fftfreq(cut_length, 1 / srate)[:cut_length // 2]
    fig, axs = plt.subplots(1, 6)
    for ind in range(6):
        A_qseis = np.abs(np.fft.fft(qseis_list[ind]))[:cut_length // 2] / srate
        A_qssp = np.abs(np.fft.fft(qssp_list[ind]))[:cut_length // 2] / srate
        axs[ind].plot(f, A_qseis, marker='o', label="qseis", color="red")
        axs[ind].plot(f, A_qssp, marker='o', label="qssp", color="blue")
        axs[ind].plot(f, np.ones_like(A_qseis) * np.abs(stress_edcmp[ind]),
                      label="edcmp", color="black")
        axs[ind].set_yscale('log')
        axs[ind].set_ylabel("s_%s" % name_list[ind])
        axs[ind].set_xlabel("Frequency (Hz)")
        axs[ind].legend(loc='lower right')
        # print(A_qseis[0], np.sum(qseis_list[ind]) / srate)
    plt.subplots_adjust(wspace=0.1, hspace=0.1)
    plt.show()

    A_max = max(np.max(np.abs(qseis_list)), np.max(np.abs(qssp_list))) * 1.1

    t = np.linspace(0, cut_length / srate, cut_length, endpoint=False)
    fig, axs = plt.subplots(6, 1)
    for ind in range(6):
        # qseis_list[ind] = np.convolve(qseis_list[ind], stf)[:cut_length] / srate
        # qssp_list[ind] = np.convolve(qssp_list[ind], stf)[:cut_length] / srate
        if output_type == 'stress_rate':
            qseis_list[ind] = np.cumsum(qseis_list[ind]) / srate
            qssp_list[ind] = np.cumsum(qssp_list[ind]) / srate
        axs[ind].plot(t, qseis_list[ind], label="qseis", color="red")
        axs[ind].plot(t, qssp_list[ind], label="qssp", color="green")
        # axs[ind].plot(t, qssp_direct_list[ind], label="qssp_direct", color="blue")
        axs[ind].plot(t, np.ones_like(qseis_list[ind]) * stress_edcmp[ind],
                      label="edcmp", color="black")
        # axs[ind].set_ylim([-A_max, A_max])
        axs[ind].set_ylabel("s_%s" % name_list[ind])
    axs[-1].set_xlabel("Time (s)")
    axs[-1].legend(loc="upper right")
    plt.show()

    print(['qseis: '] + ["%10.4e" % _ for _ in np.mean(qseis_list[:, cut_length - 10: cut_length], axis=1)])
    print(['qssp : '] + ["%10.4e" % _ for _ in np.mean(qssp_list[:, cut_length - 10: cut_length], axis=1)])
    print(['edcmp: '] + ["%10.4e" % _ for _ in stress_edcmp])



from pygrnwang.create_qssp2020_bulk import *

if __name__ == '__main__':
    wavelet_duration = 5
    sampling_interval = 0.25
    time_window = 512-sampling_interval
    max_frequency = 2
    max_slowness = 0.4
    min_harmonic = 20000
    max_harmonic = 20000
    path_green = "/e/grns_test/qssp_%d_%.3f_%.3f_%.3f_%.3f_%d" % (
        wavelet_duration, sampling_interval, time_window,
        max_frequency, max_slowness, min_harmonic
    )
    # wavelet_duration = 0
    # sampling_interval = 0.5
    # time_window = 256 - sampling_interval
    # max_frequency = 0.5
    # max_slowness = 0.4
    # min_harmonic = 10000
    # path_green = "/e/grns_qssp2020/myanmar"
    os.makedirs(path_green, exist_ok=True)
    output_observables = [1 for _ in range(11)]
    pre_process_qssp2020(
        processes_num=6,
        path_green=path_green,
        path_bin='/home/zjc/python_works/pygrnwang/pygrnwang/exec/qssp2020.bin',
        # event_depth_list=list(np.linspace(2.5, 37.5, 8)),
        # receiver_depth_list=list(np.linspace(2.5, 37.5, 8)),
        # dist_range=[0, 800],
        # delta_dist=2,
        event_depth_list=[10],
        receiver_depth_list=[1],
        dist_range=[0, 100],
        delta_dist=10,
        spec_time_window=time_window,
        sampling_interval=sampling_interval,
        max_frequency=max_frequency,
        max_slowness=max_slowness,
        anti_alias=0.01,
        turning_point_filter=0,
        turning_point_d1=0,
        turning_point_d2=0,
        free_surface_filter=1,
        gravity_fc=0,
        gravity_harmonic=0,
        cal_sph=1,
        cal_tor=1,
        min_harmonic=min_harmonic,
        max_harmonic=max_harmonic,
        source_radius=0,
        source_duration=wavelet_duration*sampling_interval,
        output_observables=output_observables,
        time_window=time_window,
        time_reduction=-20,
        path_nd='/home/zjc/python_works/pygrnwang/test/ak135fc.nd',
        earth_model_layer_num=30,
        physical_dispersion=0,
        check_finished_tpts_table=False
    )
    create_grnlib_qssp2020_parallel(
        path_green=path_green, check_finished=False, cal_spec=True
    )



from pygrnwang.create_qseis2025_bulk import *

if __name__ == "__main__":
    wavelet_type = 2
    wavelet_duration = 5
    sampling_interval = 0.25
    time_window = (512 - 1) * sampling_interval
    wavenumber_sampling_rate = 12
    path_green = "/e/grns_test/qseis_%d_%d_%.3f_%.3f_%d_new_m" % (
        wavelet_type, wavelet_duration, sampling_interval, time_window,
        wavenumber_sampling_rate
    )
    # path_green = '/e/grn_qseis2025/ak135'
    pre_process_qseis2025(
        processes_num=1,
        path_green=path_green,
        path_bin="/home/zjc/python_works/pygrnwang/pygrnwang/exec/qseis2025.bin",
        event_depth_list=[10],
        receiver_depth_list=[1],
        dist_range=[1, 100],
        delta_dist=10,
        # event_depth_list=[h for h in range(1, 31, 2)],
        # receiver_depth_list=[h for h in range(0, 30, 2)],
        # dist_range=[0, 1000-1],
        # delta_dist=1,
        N_each_group=100,
        time_window=time_window,
        sampling_interval=sampling_interval,
        output_observables=[1, 1, 1, 1, 1],
        slowness_int_algorithm=0,
        eps_estimate_wavenumber=1e-6,
        source_radius_ratio=0.05,
        slowness_window=[0, 0, 0, 0],
        time_reduction_velo=0,
        wavenumber_sampling_rate=wavenumber_sampling_rate,
        anti_alias=0.01,
        free_surface=True,
        wavelet_duration=wavelet_duration,
        wavelet_type=wavelet_type,
        flat_earth_transform=False,
        path_nd='/home/zjc/python_works/pygrnwang/test/ak135fc.nd',
        earth_model_layer_num=30,
        # path_nd="/e/grns_test/test_lzw_edcmp/test.nd",
        # earth_model_layer_num=3,
    )
    create_grnlib_qseis2025_parallel(path_green=path_green, check_finished=False)



import os
import json
import math

import numpy as np
from scipy.interpolate import RegularGridInterpolator
import pandas as pd

from .geo import d2km


def read_edcmp_raw(path_green, output_type, obs_depth, obs_depth_list):
    grn_obs_depth = obs_depth_list[
        np.argmin(np.abs(obs_depth - np.array(obs_depth_list)))
    ]
    df = pd.read_csv(
        str(
            os.path.join(
                path_green, "edcmp2", "%.2f" % grn_obs_depth, "hs.%s" % output_type
            )
        ),
        skiprows=3,
        sep="\\s+",
        header=None,
    )
    # Ux_m, Uy_m, Uz_m
    # Sxx_Pa Syy_Pa Szz_Pa Sxy_Pa Syz_Pa Szx_Pa
    values_raw = df.to_numpy()[:, 2:]
    return values_raw


def interpolate_values(xmin, xmax, nx, ymin, ymax, ny, v_array, obs_array):
    x = np.linspace(xmin, xmax, nx)
    y = np.linspace(ymin, ymax, ny)
    v_grid = v_array.reshape(ny, nx, v_array.shape[1])
    interpolator = RegularGridInterpolator((y, x), v_grid)
    pts = obs_array[:, [1, 0]]
    sigma_interp = interpolator(pts)
    return sigma_interp


def seek_edcmp2(
    path_green: str,
    output_type: str,
    obs_array: np.ndarray,
    geo_coordinate=True,
):
    """
    ee en ez nn nz zz
    :param path_green: the root dir of Green's function lib
    :param output_type: 'disp','strain','stress','tilt'
    :param obs_array:
      Cartesian coordinate, [[x_1, y_1, z_1]], [x_2, y_2, z_2],..., [x_n, y_n, z_n]], or
      Geology coordinate, [[lat_1, lon_1, dep_1], [lat_2, lon_2, dep_2],
                          ..., [lat_n, lon_n, dep_n]]
      shape is (n, 3)
    :param geo_coordinate: True/False, use Geology/Cartesian coordinate

    :return values at obs_array
    """
    obs_array = obs_array.copy()
    with open(os.path.join(path_green, "green_lib_info.json"), "r") as fr:
        green_info = json.load(fr)
    obs_x_range = green_info["obs_x_range"]
    obs_y_range = green_info["obs_y_range"]
    obs_delta_x = green_info["obs_delta_x"]
    obs_delta_y = green_info["obs_delta_y"]
    obs_depth_list = green_info["obs_depth_list"]
    if not isinstance(obs_depth_list, list):
        obs_depth_list = [obs_depth_list]
    nx = math.ceil((obs_x_range[1] - obs_x_range[0]) / obs_delta_x) + 1
    ny = math.ceil((obs_y_range[1] - obs_y_range[0]) / obs_delta_y) + 1
    x_min, x_max = obs_x_range
    y_min, y_max = obs_y_range

    if output_type == "disp":
        cha_num = 3
    elif output_type == "strain":
        cha_num = 6
    elif output_type == "stress":
        cha_num = 6
    elif output_type == "tilt":
        cha_num = 2
    else:
        raise ValueError("output_type must in disp,strain,stress,tilt")
    values_output = np.zeros((obs_array.shape[0], cha_num)) + np.nan
    if geo_coordinate:
        obs_ref = green_info["obs_ref"]
        if obs_ref is None:
            raise ValueError("The Green's function lib dose not use obs_ref!")
        elif (
            np.min(obs_array[:, 0]) - obs_ref[0] < x_min / d2km
            or np.max(obs_array[:, 0]) - obs_ref[0] > x_max / d2km
            or np.min(obs_array[:, 1]) - obs_ref[1] < y_min / d2km
            or np.max(obs_array[:, 1]) - obs_ref[1] > y_max / d2km
            # or np.min(obs_array[:, 2]) < obs_depth_list[0]
            # or np.max(obs_array[:, 2]) > obs_depth_list[-1]
        ):
            print(
                np.min(obs_array[:, 0]) - obs_ref[0] < x_min / d2km,
                np.max(obs_array[:, 0]) - obs_ref[0] > x_max / d2km,
                np.min(obs_array[:, 1]) - obs_ref[1] < y_min / d2km,
                np.max(obs_array[:, 1]) - obs_ref[1] > y_max / d2km,
                # np.min(obs_array[:, 2]) < obs_depth_list[0],
                # np.max(obs_array[:, 2]) > obs_depth_list[-1]
            )
            raise ValueError("obs_array exceeds the range of Green's function lib!")
        else:
            obs_array[:, 0] = (obs_array[:, 0] - obs_ref[0]) * d2km
            obs_array[:, 1] = (obs_array[:, 1] - obs_ref[1]) * d2km

    unique_depths = np.unique(obs_array[:, 2])
    for i in range(len(unique_depths)):
        inds_i = np.where(obs_array[:, 2] == unique_depths[i])
        obs_array_dep_i = obs_array[inds_i]
        v_dep_i = read_edcmp_raw(
            path_green, output_type, unique_depths[i], obs_depth_list
        )
        v_interp_dep_i = interpolate_values(
            x_min, x_max, nx, y_min, y_max, ny, v_dep_i, obs_array_dep_i
        )
        values_output[inds_i] = v_interp_dep_i
    values_output_enz = np.zeros_like(values_output)
    # nn ee zz ne -ez -nz
    # ee en ez nn nz zz
    values_output_enz[:, 0] = values_output[:, 1]
    values_output_enz[:, 1] = values_output[:, 3]
    values_output_enz[:, 2] = -values_output[:, 4]
    values_output_enz[:, 3] = values_output[:, 0]
    values_output_enz[:, 4] = -values_output[:, 5]
    values_output_enz[:, 5] = values_output[:, 2]
    return values_output_enz



import pandas as pd
import numpy as np
from pygrnwang.read_spgrn import *
from pygrnwang.read_qssp2020 import *

from pygrnwang.obspy_taup import cal_first_p

if __name__ == '__main__':
    event_depth_km = 10
    receiver_depth_km = 1
    az_deg = 20
    dist_km = 300
    u_spgrn_list = []
    u_qssp_list = []
    for n in [1, 0, -1, -2, -3, -4]:
        sampling_interval = 2 ** n
        srate = 1 / sampling_interval
        fm = [0, 0, 1, 0, 0, 0]
        ind = 2
        tp = cal_first_p(event_depth_km,
                         dist_km,
                         receiver_depth_km,
                         model_name="/e/grns_test/test_spgrn/noq.nd")
        (
            u_spgrn,
            tpts_table,
            first_p,
            first_s,
            grn_dep_source,
            grn_dep_receiver,
            grn_dist,
        ) = seek_spgrn2020(
            path_green="/e/grns_test/test_spgrn_%f" % sampling_interval,
            event_depth_km=event_depth_km,
            receiver_depth_km=receiver_depth_km,
            az_deg=az_deg,
            dist_km=dist_km,
            focal_mechanism=fm,
            srate=srate,
            before_p=None,
            pad_zeros=False,
            shift=False,
            rotate=True,
            only_seismograms=False,
            model_name="/e/grns_test/test_spgrn/noq.nd",
            green_info=None,
        )
        u_spgrn = np.concatenate([np.zeros(round(tpts_table['p_onset'] * srate)), u_spgrn[ind]])
        u_spgrn = u_spgrn[round(20 * srate):]
        print(tpts_table)
        print(tp)
        print(grn_dep_source)
        print(grn_dep_receiver)
        print(grn_dist)
        print()

        (
            u_qssp,
            tpts_table,
            first_p,
            first_s,
            grn_dep_source,
            grn_dep_receiver,
            grn_dist,
        ) = seek_qssp2020(
            path_green="/e/grns_test/test_qssp_%f" % sampling_interval,
            event_depth_km=event_depth_km,
            receiver_depth_km=receiver_depth_km,
            az_deg=az_deg,
            dist_km=dist_km,
            focal_mechanism=fm,
            srate=srate,
            before_p=None,
            pad_zeros=False,
            shift=False,
            rotate=True,
            only_seismograms=False,
            output_type="velo",
            model_name="/e/grns_test/test_qssp/noq.nd",
            green_info=None,
        )
        u_qssp = u_qssp[ind]
        # u_qssp = np.concatenate([np.zeros(2), u_qssp])
        print(tpts_table)
        print(grn_dep_source)
        print(grn_dep_receiver)
        print(grn_dist)
        print()

        u_spgrn_list.append(u_spgrn)
        u_qssp_list.append(u_qssp)

    import matplotlib

    matplotlib.use('tkagg')

    import matplotlib.pyplot as plt
    exp_list = [1, 0, -1, -2, -3, -4]
    fig,axs=plt.subplots(6,1, figsize=(24, 10))
    for n in range(6):
        print(n, len(u_qssp_list[n]),2**exp_list[n])
        print(n, len(u_spgrn_list[n]), 2 ** exp_list[n])
        t = np.linspace(0, len(u_qssp_list[n])*(2**exp_list[n]), len(u_qssp_list[n]), endpoint=False)
        axs[n].plot(t, u_qssp_list[n], label='qssp', linewidth=0.5)
        axs[n].plot(t, u_spgrn_list[n][:len(u_qssp_list[n])], label='spgrn', linewidth=0.5)
    plt.legend()
    plt.show()

    plt.figure()
    plt.plot(u_qssp_list[4], label='qssp')
    plt.plot(u_spgrn_list[4], label='spgrn')
    plt.legend()
    plt.show()

    plt.figure()
    plt.plot(u_qssp_list[5], label='qssp')
    plt.plot(u_spgrn_list[5], label='spgrn')
    plt.legend()
    plt.show()

    # from scipy.signal import firwin, freqz, lfilter
    #
    # cutoff = 0.05
    # numtaps = 21
    # taps = firwin(numtaps, cutoff, fs=srate, window='hamming')
    # u_qssp_conv = np.convolve(u_qssp, taps, mode='full')
    # u_spgrn_conv = np.convolve(u_spgrn, taps, mode='full')
    #
    # plt.figure()
    # plt.plot(u_qssp_conv, label='qssp')
    # plt.plot(u_spgrn_conv, label='spgrn')
    # plt.legend()
    # plt.show()



s = """# This is the input file of FORTRAN77 program "qssp2020" for calculating
# synthetic seismograms of a self-gravitating, spherically symmetric,
# isotropic and viscoelastic earth.
#
# by
# Rongjiang Wang <wang@gfz-potsdam.de>
# Helmholtz-Centre Potsdam
# GFZ German Reseach Centre for Geosciences
# Telegrafenberg, D-14473 Potsdam, Germany
#
# Last modified: Potsdam, April 2020
#
# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
# If not specified, SI Unit System is used overall!
#
# Coordinate systems:
# spherical (r,t,p) with r = radial, t = co-latitude, and p = east longitude.
# local cartesian (e,n,z) with e = east, n = north, and z = vertical (upwards positve).
# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
#
#   UNIFORM RECEIVER DEPTH
#   ======================
# 1. uniform receiver depth [km]
#--------------------------------------------------------------------------------------------------------
    0.00
#--------------------------------------------------------------------------------------------------------
#
#   SPACE-TIME SAMPLING PARAMETERS
#   ==============================
# 1. time window [sec], sampling interval [sec]
# 2. max. frequency [Hz] of Green's functions
# 3. max. slowness [s/km] of Green's functions
#    Note: if the near-field static displacement is desired, the maximum slowness should not
#          be smaller than the S wave slowness in the receiver layer
# 4. anti-aliasing factor (> 0 & < 1), if it is <= 0 or >= 1/e (~ 0.4), then default value of
#    1/e is used (e.g., 0.1 = alias phases will be suppressed to 10% of their original
#    amplitude)
# 5. switch (1/0 = yes/no) of turning-point filter, the range (d1, d2) of max. penetration
#    depth [km] (d1 is meaningless if it is smaller than the receiver/source depth, and
#    d2 is meaningless if it is equal to or larger than the earth radius)
#
#    Note: The turning-point filter (Line 5) works only for the extended QSSP code (e.g.,
#          qssp2016). if this filter is selected, all phases with the turning point
#          shallower than d1 or deeper than d2 will be filtered.
#
# 6. Earth radius [km], switch of free-surface-reflection filter (1/0 = with/without free
#    surface reflection)
#
#    Note: The free-surface-reflection filter (Line 6) works only for the extended QSSP
#          code (e.g., qssp2016). if this filter is selected, all phases with the turning
#          point shallower than d1 or deeper than d2 will be filtered.
#--------------------------------------------------------------------------------------------------------
    255.0   1.00
    0.05
    0.40
    0.01
	0   2891.5   6371.0
	6371.0    1
#--------------------------------------------------------------------------------------------------------
#
#   SELF-GRAVITATING EFFECT
#   =======================
# 1. the critical frequency [Hz] and the critical harmonic degree, below which
#    the self-gravitating effect should be included
#--------------------------------------------------------------------------------------------------------
    0.05     500
#--------------------------------------------------------------------------------------------------------
#
#   WAVE TYPES
#   ==========
# 1. selection (1/0 = yes/no) of spheroidal modes (P-SV waves), selection of toroidal modes
#    (SH waves), minimum and maximum cutoff harmonic degrees
#    Note: if the near-field static displacement is desired, the minimum cutoff harmonic
#          degree should not be smaller than, e.g., 2000.
#--------------------------------------------------------------------------------------------------------
    1     1    20000  20000
#--------------------------------------------------------------------------------------------------------
#   GREEN'S FUNCTION FILES
#   ======================
# 1. number of discrete source depths, estimated radius of each source patch [km] and
#    directory for Green's functions
# 2. list of the source depths [km], the respective file names of the Green's functions
#    (spectra) and the switch number (0/1) (0 = do not calculate this Green's function because
#    it exists already, 1 = calculate or update this Green's function).
#    Note: Green's functions need to be recalculated if any of the above parameters is changed.
#--------------------------------------------------------------------------------------------------------
    1     0.0  '/home/zjc/Desktop/test_qssp_gf/'
    0.00   'Green_0km'    0
#--------------------------------------------------------------------------------------------------------
#
#   MULTI-EVENT SOURCE PARAMETERS
#   =============================
# 1. number of discrete point sources and selection of the source data format
#    (1, 2 or 3)
# 2. list of the multi-event sources
#
#    Format 1 (full moment tensor):
#    Unit     Mrr  Mtt  Mpp  Mrt  Mrp  Mtp  Lat   Lon   Depth  T_origin T_rise
#    [Nm]                                   [deg] [deg] [km]   [sec]    [sec]
#
#    Format 2 (double couple):
#    Unit   Strike    Dip       Rake      Lat   Lon   Depth  T_origin T_rise
#    [Nm]   [deg]     [deg]     [deg]     [deg] [deg] [km]   [sec]    [sec]
#
#    Format 3 (single force):
#    Unit      Feast    Fnorth  Fvertical    Lat   Lon   Depth  T_origin T_rise
#    [N]                                     [deg] [deg] [km]   [sec]    [sec]
#
#    Note: for each point source, the default moment (force) rate time function is used, defined by a
#          squared half-period (T_rise) sinusoid starting at T_origin.
#--------------------------------------------------------------------------------------------------------
    1      1
    1  1.0 0.0 0.0 0.0 0.0 0.0  0.0 0.0 10.0   0.0 0.0
#--------------------------------------------------------------------------------------------------------
#
#   RECEIVER PARAMETERS
#   ===================
# 1. select output observables (1/0 = yes/no)
#    Note: "gravitation" here means (space-based) garvitational acceleration vector, while "gravity"
#          means the gravity change (downwards positive) measured by a gravimeter. The latter includes
#          the effect due to free-air gradient and ground acceleration.
# 2. output file name
# 3. output time window [sec] (<= Green's function time window)
# 4. selection of order of Butterworth bandpass filter (if <= 0, then no filtering), lower
#    and upper corner frequencies (<= cut-off frequency defined above)
# 5. lower and upper slowness cut-off [s/km] (slowness band-pass filter)
# 6. number of receiver
# 7. list of the station parameters
#    Format:
#    Lat     Lon    Name     Time_reduction
#    [deg]   [deg]           [sec]
#    (Note: Time_reduction = start time of the time window)
#--------------------------------------------------------------------------------------------------------------------
# disp | velo | acce | strain | strain_rate | stress | stress_rate | rotation | rotation_rate |gravitation |  gravity
#--------------------------------------------------------------------------------------------------------------------
   1      1      1      1              1        1        1            0        0            0      0
  '/home/zjc/Desktop/test_qssp_gf_output/test'
  256.0
  0   0.05   0.25
  0.00   0.40
  1
#--------------------------------------------------------------------------------------------------------
#
#   MULTI-LAYERED EARTH MODEL
#   =========================
# 1. number of data lines of the layered model and selection for including
#    the physical dispersion according to Kamamori & Anderson (1977)
#--------------------------------------------------------------------------------------------------------
    10     0
#--------------------------------------------------------------------------------------------------------
#
#   MODEL PARAMETERS
#   ================
# no   depth[km]   vp[km/s]   vs[km/s]   ro[g/cm^3] qp         qs
#--------------------------------------------------------------------------------------------------------
  1      0.000     5.8000     3.2000     2.6000    1478.30     599.99
  2      3.300     5.8000     3.2000     2.6000    1478.30     599.99
  3      3.300     5.8000     3.2000     2.6000    1478.30     599.99
  4     10.000     5.8000     3.2000     2.6000    1478.30     599.99
  5     10.000     6.8000     3.9000     2.9200    1368.02     599.99
  6     18.000     6.8000     3.9000     2.9200    1368.02     599.99
  7     18.000     8.0355     4.4839     3.6410     950.50     394.62
  8     43.000     8.0379     4.4856     3.5801     972.77     403.93
  9     80.000     8.0400     4.4800     3.5020    1008.71     417.59
 10     80.000     8.0450     4.4900     3.5020     182.03      75.60
#---------------------------------end of all inputs------------------------------------------------------"""
